"""Utility functions to format sqlalchemy query data as JSON API."""
from sqlalchemy.orm import Query


def query_data_pagination(query: Query, page: int = 1, page_size: int = 50):
    """
        Format query data with pagination
    """
    page = max(1, page)  # min page number is 1
    page_size = min(1000, page_size)  # max page size is 1000
    offset = (page - 1) * page_size

    count = query.count()
    results = [u._asdict() for u in query.offset(offset).limit(page_size).all()]
    pcount = int(count / page_size)
    pcount += 0 if (count % page_size) == 0 else 1

    return {
        'data': results,
        'meta': {
            'pagination': {
                'count': count,
                'page': page,
                'page_count': pcount,
                'page_items': len(results),
                'page_size': page_size,
            },
        }
    }



"""Utilities for handling types"""
from importlib import import_module


def load_type(name: str, parent_type: type | None = None, default_module:
              str | None = None) -> type:
    """
        Return a type from a string with a qualified or unqualified name.

        Args:
            name: The name of the class to load, can be a fully qualified name
                or an unqualified name.
            parent_type: The parent type that the class must extend.
                No check is performed if None is passed.
            default_module: The default module to use if the name is unqualified.
                Required if the name is unqualified.
    """
    if '.' not in name:  # unqualified name
        if not default_module:
            raise ValueError('default_module is needed for an unqualified name')
        name = f'{default_module}.{name}'
    module_name, class_name = name.rsplit('.', 1)
    try:
        module = import_module(module_name)
    except ModuleNotFoundError:
        raise ValueError(f'Module "{module_name}" not found')
    if not hasattr(module, class_name):
        raise ValueError(f'Class "{class_name}" not found in "{module_name}"')
    new_class = getattr(module, class_name)
    if parent_type and not issubclass(new_class, parent_type):
        raise ValueError(f'Class "{class_name}" must extend "{parent_type.__name__}"')

    return new_class



""" Event handling for Brevia. """
from fastapi import FastAPI

from brevia.providers import update_providers


def app_events(app: FastAPI) -> None:
    """Add event handlers to FastAPI instance."""
    app.add_event_handler('startup', update_providers)



"""Chat history table & utilities"""
from typing import List
import logging
from langchain_community.vectorstores.pgembedding import BaseModel, CollectionStore
from pydantic import BaseModel as PydanticModel
import sqlalchemy
from sqlalchemy import func
from sqlalchemy.dialects.postgresql import JSON, UUID
from sqlalchemy.orm import Mapped, Query, Session
from sqlalchemy.sql.expression import BinaryExpression
from brevia.connection import db_connection
from brevia.models import load_embeddings
from brevia.settings import get_settings
from brevia.utilities.dates import date_filter
from brevia.utilities.json_api import query_data_pagination
from brevia.utilities.uuid import is_valid_uuid


class ChatHistoryStore(BaseModel):
    # pylint: disable=too-few-public-methods
    """ Chat History table """
    __tablename__ = "chat_history"

    session_id = sqlalchemy.Column(UUID(as_uuid=True))
    collection_id: Mapped[UUID] = sqlalchemy.Column(
        UUID(as_uuid=True),
        sqlalchemy.ForeignKey(
            f"{CollectionStore.__tablename__}.uuid",
            ondelete="CASCADE",
        ),
    )
    question = sqlalchemy.Column(sqlalchemy.String)
    answer = sqlalchemy.Column(sqlalchemy.String)
    # pylint: disable=not-callable
    created = sqlalchemy.Column(
        sqlalchemy.DateTime(timezone=False),
        nullable=False,
        server_default=sqlalchemy.sql.func.now()
    )
    cmetadata = sqlalchemy.Column(JSON, nullable=True)
    user_evaluation = sqlalchemy.Column(
        sqlalchemy.BOOLEAN(),
        nullable=True,
    )
    user_feedback = sqlalchemy.Column(sqlalchemy.String)
    chat_source = sqlalchemy.Column(sqlalchemy.String)


def history(chat_history: list, session: str = None):
    """ Load chat history from input or from DB """
    if len(chat_history) > 0:
        return [(x["query"], x["answer"]) for x in chat_history]

    if not is_valid_uuid(session) or get_settings().qa_no_chat_history:
        return []

    return history_from_db(session)


def is_related(chat_history: list, question: str, embeddings: dict | None = None):
    """
    Determine whether a question is related to a sequence of sentences.
    Use sentence and question embeddings to calculate the product
    scale between the vectors (in this case = similarity) and compare it
    with a threshold specified by environment variables.
    """
    embeddings_engine = load_embeddings(embeddings)
    q_e = embeddings_engine.embed_query(question)
    h_e = embeddings_engine.embed_query(
        ''.join([sentence for tuple in chat_history for sentence in tuple])
    )
    sim = dot_product(q_e, h_e)
    logging.getLogger(__name__).info("similarity: %s", sim)
    threshold = get_settings().qa_followup_sim_threshold
    return sim >= threshold


def dot_product(v1_list, v2_list):
    """
    Calculate the scalar product between two vectors (similarity).
    """
    return sum(x * y for x, y in zip(v1_list, v2_list))


def history_from_db(session_id: str) -> List[tuple[str, str]]:
    """ Load chat history from DB """
    connection = db_connection()

    filter_by = ChatHistoryStore.session_id == session_id
    with Session(connection) as session:
        results: List[ChatHistoryStore] = (
            session.query(ChatHistoryStore)
            .filter(filter_by)
            .order_by(sqlalchemy.desc(ChatHistoryStore.created))
            .limit(3)
            .all()
        )

    chat_hist = [(x.question, x.answer) for x in results]
    chat_hist.reverse()

    return chat_hist


def add_history(
    session_id: str,
    collection: str,
    question: str,
    answer: str,
    metadata: dict | None = None,
    chat_source: str | None = None,
) -> (ChatHistoryStore | None):
    """Save chat history item to database """
    if not is_valid_uuid(session_id):
        return None

    with Session(db_connection()) as session:
        collection_uuid = None  # Default empty UUID
        if collection:
            collection_store = CollectionStore.get_by_name(session, collection)
            collection_uuid = collection_store.uuid if collection_store else None

        chat_history_store = ChatHistoryStore(
            session_id=session_id,
            collection_id=collection_uuid,
            question=question,
            answer=answer,
            cmetadata=metadata,
            chat_source=chat_source,
        )
        session.expire_on_commit = False
        session.add(chat_history_store)
        session.commit()

        return chat_history_store


class ChatHistoryFilter(PydanticModel):
    """ Chat history filter """
    min_date: str | None = None
    max_date: str | None = None
    collection: str | None = None
    session_id: str | None = None
    page: int = 1
    page_size: int = 50


def get_collection_filter(collection_name):
    """
    Constructs a filter expression based on the collection name.
    """
    filter_collection = CollectionStore.name == collection_name
    if collection_name is None:
        filter_collection = CollectionStore.name is not None
    return filter_collection


def get_session_filter(session_id):
    """
    Constructs a filter expression based on the session ID.
    """
    filter_session_id = sqlalchemy.text('1 = 1')  # (default) always true expression
    if session_id and is_valid_uuid(session_id):
        filter_session_id = ChatHistoryStore.session_id == session_id

    return filter_session_id


def get_history(filter: ChatHistoryFilter) -> dict:  # pylint: disable=redefined-builtin
    """
    Read chat history with optional filters using pagination data in response.
    """

    min_date = date_filter(filter.min_date, 'min')
    max_date = date_filter(filter.max_date, 'max')
    filter_collection = get_collection_filter(filter.collection)
    filter_session_id = get_session_filter(filter.session_id)

    with Session(db_connection()) as session:
        query = get_history_query(
            session=session,
            filter_min_date=ChatHistoryStore.created >= min_date,
            filter_max_date=ChatHistoryStore.created <= max_date,
            filter_collection=filter_collection,
            filter_session_id=filter_session_id,
        )
        result = query_data_pagination(
            query=query,
            page=filter.page,
            page_size=filter.page_size
        )
        return result


def get_history_sessions(filter: ChatHistoryFilter) -> dict:
    # pylint: disable=redefined-builtin
    """
    Read chat history with optional filters using pagination data in response.
    """

    min_date = date_filter(filter.min_date, 'min')
    max_date = date_filter(filter.max_date, 'max')
    filter_collection = get_collection_filter(filter.collection)

    with Session(db_connection()) as session:
        query = get_history_sessions_query(
            session=session,
            filter_min_date=ChatHistoryStore.created >= min_date,
            filter_max_date=ChatHistoryStore.created <= max_date,
            filter_collection=filter_collection,
        )
        result = query_data_pagination(
            query=query,
            page=filter.page,
            page_size=filter.page_size
        )
        return result


def get_history_sessions_query(
    session: Session,
    filter_min_date: BinaryExpression,
    filter_max_date: BinaryExpression,
    filter_collection: BinaryExpression,
) -> Query:
    """
    Returns a SQLAlchemy query to fetch oldest chat history sessions
    for every session based on specified filters.
    """

    subquery = (
        session.query(
            ChatHistoryStore.session_id.label("session_id"),
            CollectionStore.name.label("collection"),
            # pylint: disable=not-callable
            func.min(ChatHistoryStore.created).label("min_created"),
        ).join(
            CollectionStore,
            CollectionStore.uuid == ChatHistoryStore.collection_id
        )
        .filter(filter_min_date, filter_max_date, filter_collection)
        .group_by(
            ChatHistoryStore.session_id,
            CollectionStore.name
        )
    ).subquery()

    query = (
        session.query(
            ChatHistoryStore.session_id,
            ChatHistoryStore.question,
            ChatHistoryStore.created,
            subquery.c.collection,
        )
        .join(subquery, ChatHistoryStore.session_id == subquery.c.session_id)
        .filter(ChatHistoryStore.created == subquery.c.min_created)
        .order_by(sqlalchemy.desc(ChatHistoryStore.created))
    )

    return query


def get_history_query(
    session: Session,
    filter_min_date: BinaryExpression,
    filter_max_date: BinaryExpression,
    filter_collection: BinaryExpression,
    filter_session_id: BinaryExpression,
) -> Query:
    """
    Constructs a SQLAlchemy query to retrieve chat history based on specified filters.
    """

    query = (
        session.query(
            ChatHistoryStore.uuid,
            ChatHistoryStore.question,
            ChatHistoryStore.answer,
            ChatHistoryStore.session_id,
            ChatHistoryStore.cmetadata,
            ChatHistoryStore.created,
            CollectionStore.name.label("collection"),
            ChatHistoryStore.user_evaluation,
            ChatHistoryStore.user_feedback,
            ChatHistoryStore.chat_source,
        )
        .join(
            CollectionStore,
            CollectionStore.uuid == ChatHistoryStore.collection_id
        )
        .filter(filter_min_date, filter_max_date, filter_collection, filter_session_id)
        .order_by(sqlalchemy.desc(ChatHistoryStore.created))
    )
    return query


def history_evaluation(
    history_id: str,
    user_evaluation: bool,
    user_feedback: str | None = None,
    metadata: dict | None = None,
) -> bool:
    """
        Update evaluation of single history item.
        Return false if history item is not found and true upon success.
    """
    with Session(db_connection()) as session:
        chat_history = session.get(ChatHistoryStore, history_id)
        if chat_history is None:
            return False
        chat_history.user_evaluation = user_evaluation
        chat_history.user_feedback = user_feedback
        if metadata:
            if chat_history.cmetadata is None:
                chat_history.cmetadata = {}
            chat_history.cmetadata = chat_history.cmetadata | metadata

        session.add(chat_history)
        session.commit()

    return True



"""DB migrations commands """
from os import getcwd
from os.path import dirname
from alembic.config import Config
from alembic import command

alembic_cfg = Config(f'{getcwd()}/alembic.ini')
alembic_cfg.set_main_option('script_location', dirname(__file__))


def current(verbose=False):
    command.current(alembic_cfg, verbose=verbose)


def upgrade(revision="head"):
    command.upgrade(alembic_cfg, revision)


def downgrade(revision):
    command.downgrade(alembic_cfg, revision)



"""config table

Revision ID: 24b025d48e0a
Revises: 95d93297832c
Create Date: 2024-10-11 10:19:27.308907

"""
from alembic import op
import sqlalchemy as sa
import uuid

# revision identifiers, used by Alembic.
revision = '24b025d48e0a'
down_revision = '95d93297832c'
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        'config',
        sa.Column('uuid', sa.UUID(), primary_key=True, default=uuid.uuid4),
        sa.Column(
            'config_key',
            sa.String(),
            nullable=False,
            comment='Configuration name'
        ),
        sa.Column(
            'config_val',
            sa.String(),
            nullable=False,
            comment='Configuration value'
        ),
        sa.Column(
            'created',
            sa.TIMESTAMP(timezone=False),
            nullable=False,
            server_default=sa.func.current_timestamp(),
            comment='Creation timestamp'
        ),
        sa.Column(
            'modified',
            sa.TIMESTAMP(timezone=False),
            nullable=False,
            server_default=sa.func.current_timestamp(),
            onupdate=sa.func.current_timestamp(),
            comment='Last update timestamp'
        ),
    )


def downgrade() -> None:
    op.drop_table('config')



"""Settings module"""
import logging
from functools import lru_cache
from typing import Annotated, Any
from os import environ, path, getcwd
from urllib import parse
from sqlalchemy import NullPool, create_engine, Column, String, func, inspect
from sqlalchemy.engine import Connection
from sqlalchemy.dialects.postgresql import TIMESTAMP
from sqlalchemy.orm import Session
from langchain_community.vectorstores.pgembedding import BaseModel
from langchain.globals import set_verbose
from pydantic import Field, Json, PrivateAttr, ValidationError
from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    """Brevia settings"""
    model_config = SettingsConfigDict(
        env_file='.env', extra='ignore'
    )
    _defaults: dict[str, Any] = PrivateAttr(default={})

    verbose_mode: bool = False

    # Postgres+pgvector vector db
    pgvector_host: str = Field(default='localhost', exclude=True)
    pgvector_driver: str = Field(default='psycopg2', exclude=True)
    pgvector_port: int = Field(default=5432, exclude=True)
    pgvector_database: str = Field(default='brevia', exclude=True)
    pgvector_user: str = Field(default='', exclude=True)
    pgvector_password: str = Field(default='', exclude=True)
    pgvector_pool_size: int = Field(default=10, exclude=True)
    # optional DSN URI, if set other pgvector_* settings are ignored
    pgvector_dsn_uri: str = Field(default='', exclude=True)

    # Tokens
    tokens_secret: str = Field(default='', exclude=True)
    tokens_users: str = Field(default='', exclude=True)
    status_token: str = Field(default='', exclude=True)

    # External services API keys (deprecated)
    # You should use `brevia_env_secrets` to store secrets that
    # must be available as environment variables
    openai_api_key: Annotated[
        str,
        Field(deprecated='Use `brevia_env_secrets` instead')
    ] = Field(default='', exclude=True)
    cohere_api_key: Annotated[
        str,
        Field(deprecated='Use `brevia_env_secrets` instead')
    ] = Field(default='', exclude=True)

    # Test models - only in unit tests
    use_test_models: bool = Field(default=False, exclude=True)

    # Index - text splitter settings
    text_chunk_size: int = 2000
    text_chunk_overlap: int = 200
    text_splitter: Json[dict[str, Any]] = '{}'  # custom splitter settings

    # Search
    search_docs_num: int = 4

    # LLM settings
    qa_completion_llm: Json[dict[str, Any]] = """{
        "model": "gpt-4o-mini",
        "model_provider": "openai",
        "temperature": 0,
        "max_tokens": 2000
    }"""
    qa_followup_llm: Json[dict[str, Any]] = """{
        "model": "gpt-4o-mini",
        "model_provider": "openai",
        "temperature": 0,
        "max_tokens": 500
    }"""
    summarize_llm: Json[dict[str, Any]] = """{
        "model": "gpt-4o-mini",
        "model_provider": "openai",
        "temperature": 0,
        "max_tokens": 2000
    }"""

    # Environment secrets, dictionary of variables that must be present as environment
    # variables
    # (Pydantic throws an error using SecretStr as dict value type)
    brevia_env_secrets: Json[dict[str, str]] = '{}'

    # Embeddings
    embeddings: Json = '{"_type": "openai-embeddings"}'
    # every vendor has its own embeddings vector size, some vendors have multiple sizes
    # this value should be unique in your project to avoid calculation errors
    embeddings_size: Annotated[
        int,
        Field(deprecated='Define size in `embeddings` conf')
    ] = Field(default=1536, exclude=True)

    # QA
    qa_no_chat_history: bool = False  # don't load chat history
    qa_followup_sim_threshold: float = 0.735  # similitude threshold in followup
    qa_retriever: Json[dict[str, Any]] = '{}'  # custom retriever settings

    # Summarization
    summ_default_chain: str = 'stuff'
    summ_token_splitter: int = 4000
    summ_token_overlap: int = 500

    # Prompt files base path - local or shared file system
    prompts_base_path: str = Field(
        default=f'{path.dirname(__file__)}/prompts',
        exclude=True,
    )

    # Providers: list of available providers and models
    providers: Json = '[]'

    # List of known providers environment variables that could be used by
    # external services (mainly LLM providers)
    providers_env_vars: Json[dict[str, list[str]]] = """{
        "openai": ["OPENAI_API_KEY","OPENAI_ORG_ID","OPENAI_API_BASE"],
        "anthropic": ["ANTHROPIC_API_KEY"],
        "cohere": ["COHERE_API_KEY"],
        "deepseek": ["DEEPSEEK_API_KEY"],
        "ollama": ["OLLAMA_HOST"]
    }"""

    # App metadata
    block_openapi_urls: bool = False

    # File output
    file_output_base_path: str = Field(
        default=f'{path.abspath(getcwd())}/files',
        exclude=True
    )
    file_output_base_url: str = Field(
        default='/download',
        exclude=True
    )

    def update(
        self,
        other: dict[str, Any] | BaseSettings,
    ) -> None:
        """Update settings fields, used when updating from db and in unit tests"""
        keys = Settings.model_fields.keys()
        if not isinstance(other, BaseSettings):
            other = {k.lower(): v for k, v in other.items() if hasattr(self, k)}
            keys = other.keys()
            other = Settings(**other)
        for key in keys:
            newattr = getattr(other, key)
            if key == 'brevia_env_secrets':
                current_secrets = getattr(self, key)
                newattr = {**current_secrets, **newattr}
            setattr(self, key, newattr)

    def setup_environment(self):
        """Setup some useful environment variables from `brevia_env_secrets`"""
        log = logging.getLogger(__name__)
        for key, value in self.brevia_env_secrets.items():
            environ[key] = value
            log.info('"%s" env var set', key)
        for provider in self.providers_env_vars:
            for key in self.providers_env_vars[provider]:
                if key in environ and key not in self.brevia_env_secrets.keys():
                    self.brevia_env_secrets[key] = environ[key]
                    log.info('"%s" added to env secrets', key)

    def connection_string(self) -> str:
        """ Db connection string from Settings """
        if self.pgvector_dsn_uri:
            return self.pgvector_dsn_uri

        driver = self.pgvector_driver
        host = self.pgvector_host
        port = self.pgvector_port
        database = self.pgvector_database
        user = self.pgvector_user
        password = parse.quote_plus(self.pgvector_password)

        return f"postgresql+{driver}://{user}:{password}@{host}:{port}/{database}"

    def update_from_db(self):
        """Update settings from db"""
        try:
            engine = create_engine(self.connection_string(), poolclass=NullPool)
            insp = inspect(engine)
            if insp.has_table(ConfigStore.__tablename__):
                db_conf = read_db_conf(engine.connect())
                self.update(db_conf)
            engine.dispose()
        except Exception as exc:  # pylint: disable=broad-exception-caught
            logging.getLogger(__name__).error('Failed to read config from db: %s', exc)

    def setup_defaults(self):
        """Setup defaults"""
        self._defaults = self.model_dump()

    def get_configurable_schema(self) -> dict[str, Any]:
        """Get JSON Schema of configurable properties, with actual defaults"""
        schema = self.model_json_schema(mode='serialization')
        for key, value in schema['properties'].items():
            value['default'] = self._defaults[key]

        return schema


class ConfigStore(BaseModel):
    # pylint: disable=too-few-public-methods,not-callable
    """ Config table """
    __tablename__ = "config"

    config_key = Column(String(), nullable=False, unique=True)
    config_val = Column(String(), nullable=False)
    created = Column(
        TIMESTAMP(timezone=False),
        nullable=False,
        server_default=func.current_timestamp(),
    )
    modified = Column(
        TIMESTAMP(timezone=False),
        nullable=False,
        server_default=func.current_timestamp(),
        onupdate=func.current_timestamp(),
    )


@lru_cache
def get_configurable_keys():
    """Get configurable keys"""
    return Settings().model_dump().keys()


def read_db_conf(connection: Connection) -> dict[str, str]:
    """ Read all config records """
    with Session(connection) as session:
        query = session.query(ConfigStore.config_key, ConfigStore.config_val)
        # Filter out non-configurable settings
        keys = get_configurable_keys()
        return {key: value for key, value in query.all() if key in keys}


def update_db_conf(connection: Connection, items: dict[str, str]) -> dict[str, str]:
    """ Update some config items """
    # Filter out non-configurable settings
    keys = get_configurable_keys()
    items = {k: v for k, v in items.items() if k in keys}
    # validate items - this will raise a validation Error a key is not valid
    try:
        Settings(**items)
    except ValidationError as exc:
        logging.getLogger(__name__).error('Invalid settings: %s', exc)
        raise exc

    with Session(connection) as session:
        session.expire_on_commit = False
        query = session.query(ConfigStore)
        current_conf = {u.config_key: u for u in query.all()}
        for key, value in items.items():
            if key not in current_conf:
                session.add(ConfigStore(config_key=key, config_val=value))
            elif current_conf[key].config_val != value:
                current_conf[key].config_val = value
                session.add(current_conf[key])

        session.commit()
    # Remove from environment variables not in the new brevia_env_secrets
    if 'brevia_env_secrets' in items:
        new_secrets = items['brevia_env_secrets']
        for key in get_settings().brevia_env_secrets:
            if key not in new_secrets:
                del environ[key]
    # Clear settings cache, force settings reload
    get_settings.cache_clear()

    return read_db_conf(connection)


def reset_db_conf(connection: Connection, items: list[str]) -> dict[str, str]:
    """ Reset to defatul some config items """
    # Filter out non-configurable settings
    keys = get_configurable_keys()
    items = [k for k in items if k in keys]
    with Session(connection) as session:
        session.expire_on_commit = False
        query = session.query(ConfigStore)
        current_conf = {u.config_key: u for u in query.all()}
        for key in items:
            if key in current_conf:
                session.delete(current_conf[key])
        session.commit()
    # Clear settings cache, force settings reload
    get_settings.cache_clear()

    return read_db_conf(connection)


@lru_cache
def get_settings():
    """Return Settings object instance just once (using lru_cache)"""
    settings = Settings()
    settings.setup_defaults()
    settings.update_from_db()
    settings.setup_environment()

    set_verbose(settings.verbose_mode)

    return settings



"""Question-answering and search functions against a vector database."""
from langchain.retrievers.multi_query import MultiQueryRetriever
from langchain.chains.base import Chain
from langchain.chains.combine_documents import create_stuff_documents_chain
from langchain_community.vectorstores.pgembedding import CollectionStore
from langchain_community.vectorstores.pgvector import DistanceStrategy, PGVector
from langchain_core.callbacks import BaseCallbackHandler
from langchain_core.prompts import PromptTemplate
from langchain_core.output_parsers import StrOutputParser
from langchain_core.runnables import RunnablePassthrough
from langchain_core.retrievers import BaseRetriever
from langchain_core.vectorstores import VectorStore
from langchain_core.language_models import BaseChatModel
from langchain_core.documents import Document
from pydantic import BaseModel
from brevia.connection import connection_string
from brevia.collections import single_collection_by_name
from brevia.models import load_chatmodel, load_embeddings, get_model_config
from brevia.prompts import load_qa_prompt, load_condense_prompt
from brevia.settings import get_settings
from brevia.utilities.types import load_type
from brevia.base_retriever import BreviaBaseRetriever


DISTANCE_MAP = {
    'euclidean': DistanceStrategy.EUCLIDEAN,
    'cosine': DistanceStrategy.COSINE,
    'max': DistanceStrategy.MAX_INNER_PRODUCT,
}


class SearchQuery(BaseModel):
    """
    Parameters for a vector-based search query.

    Attributes:
        query (str): The search query text.
        collection (str): The collection name or identifier.
        docs_num (int | None): Max number of documents to retrieve.
        distance_strategy_name (str): Distance strategy, defaults to 'cosine'.
        filter (dict[str, str | dict | list] | None): Optional filter criteria.
    """
    query: str
    collection: str
    docs_num: int | None = None
    distance_strategy_name: str = 'cosine'
    filter: dict[str, str | dict | list] | None = None


class ChatParams(BaseModel):
    """
    Parameters for initiating a Q&A conversation chain.

    Attributes:
        docs_num (int | None): Optional number of documents to retrieve.
        streaming (bool): Flag to determine if the response should be streamed.
        distance_strategy_name (str | None): Optional strategy for distance calculation.
        filter (dict[str, str | dict] | None): Optional filter parameters for retrieval.
        source_docs (bool): Flag to include retrieved source documents in the response.
        multiquery (bool): Flag for executing multiple queries for retrieval.
        search_type (str): Type of search algorithm (default is "similarity").
        score_threshold (float): Threshold for filtering documents by relevance scores.
    """
    docs_num: int | None = None
    streaming: bool = False
    distance_strategy_name: str | None = None
    filter: dict[str, str | dict] | None = None
    source_docs: bool = False
    multiquery: bool = False
    config: dict | None = None
    search_type: str = "similarity"
    score_threshold: float = 0.0

    def get_search_kwargs(self) -> dict:
        """ Construct and return keyword arguments needed for search methods. """
        return {
            'k': self.docs_num,
            'filter': self.filter,
            'score_threshold': self.score_threshold,
        }


def search_vector_qa(
    search: SearchQuery,
) -> list[tuple[Document, float]]:
    """
    Execute a vector search for Q&A tasks and return matching documents with scores.

    This function uses the provided search parameters to perform a similarity search
    on the designated collection's vector index. It retrieves a list of document-score
    tuples that best match the input query.
    """
    collection_store = single_collection_by_name(search.collection)
    if not collection_store:
        raise ValueError(f'Collection not found: {search.collection}')
    if search.docs_num is None:
        default_num = get_settings().search_docs_num
        search.docs_num = int(collection_store.cmetadata.get('docs_num', default_num))
    strategy = DISTANCE_MAP.get(search.distance_strategy_name, DistanceStrategy.COSINE)
    embeddings_conf = collection_store.cmetadata.get('embeddings', None)
    docsearch = PGVector(
        connection_string=connection_string(),
        embedding_function=load_embeddings(embeddings_conf),
        collection_name=search.collection,
        distance_strategy=strategy,
        use_jsonb=True,
    )

    return docsearch.similarity_search_with_score(
        query=search.query,
        k=search.docs_num,
        filter=search.filter,
    )


def create_custom_retriever(
        store: VectorStore,
        search_kwargs: dict,
        retriever_conf: dict,
) -> BaseRetriever:
    """Create sda custom retriever from a configuration."""
    retriever_name = retriever_conf.pop('retriever', '')
    retriever_class = load_type(retriever_name, BaseRetriever)

    return retriever_class(
        vectorstore=store,
        search_kwargs=search_kwargs,
        **retriever_conf,
    )


def create_default_retriever(
        store: VectorStore,
        search_kwargs: dict,
        llm: BaseChatModel,
        search_type: str | None = None,
        multiquery: bool = False,

) -> BaseRetriever:
    """
        Create a default retriever.
        Can be a vector store retriever or a multiquery retriever.
    """
    retriever = BreviaBaseRetriever(
        vectorstore=store,
        search_type=search_type,
        search_kwargs=search_kwargs
    )

    if multiquery:
        return MultiQueryRetriever.from_llm(retriever=retriever, llm=llm)

    return retriever


def create_conversation_retriever(
    collection: CollectionStore,
    chat_params: ChatParams,
    llm: BaseChatModel,
) -> BaseRetriever:
    """ Create a retriever for a collection with chat parameters """
    strategy = DISTANCE_MAP.get(
        chat_params.distance_strategy_name,
        DistanceStrategy.COSINE
    )

    embeddings_conf = collection.cmetadata.get('embeddings', None)
    document_search = PGVector(
        connection_string=connection_string(),
        embedding_function=load_embeddings(embeddings_conf),
        collection_name=collection.name,
        distance_strategy=strategy,
        use_jsonb=True,
    )
    search_kwargs = chat_params.get_search_kwargs()
    retriever_conf = collection.cmetadata.get(
        'qa_retriever',
        dict(get_settings().qa_retriever).copy()
    )
    if not retriever_conf:
        return create_default_retriever(
            store=document_search,
            search_kwargs=search_kwargs,
            llm=llm,
            search_type=chat_params.search_type,
            multiquery=chat_params.multiquery,
        )

    # custom retriever
    return create_custom_retriever(
        document_search, search_kwargs, retriever_conf)


def conversation_rag_chain(
    collection: CollectionStore,
    chat_params: ChatParams,
    answer_callbacks: list[BaseCallbackHandler] | None = None,
) -> Chain:
    """
    Create and return a conversation chain for Q&A with embedded dataset knowledge.(RAG)

    Args:
        collection (CollectionStore): The collection store item containing the dataset.
        chat_params (ChatParams): Parameters for configuring the conversation chain,
            including:
            - docs_num (int | None): Number of documents to retrieve for context
              (default is from settings or collection metadata).
            - streaming (bool): Flag to enable or disable streaming responses
              (default is False).
            - distance_strategy_name (str | None): Name of the distance strategy to use
              (default is 'cosine').
            - filter (dict[str, str | dict] | None): Optional dictionary of metadata to
              use as a filter (default is None).
            - source_docs (bool): Flag to include source documents in the response
              (default is False).
            - multiquery (bool): Flag to enable multiple queries for retrieval
              (default is False).
            - search_type (str): Type of search algorithm to use (def is 'similarity').
            - score_threshold (float): Threshold for filtering documents based on
              relevance scores (default is 0.0).
            - config (dict | None): Optional configuration dict that can contain
              completion_llm and followup_llm configs to override defaults.
        answer_callbacks (list[BaseCallbackHandler] | None): List of callback handlers
            for the final LLM answer to enable streaming (default is None).

    Returns:
        Chain: A configured conversation chain for Q&A tasks.
    """
    settings = get_settings()
    if chat_params.docs_num is None:
        chat_params.docs_num = int(
            collection.cmetadata.get('docs_num', settings.search_docs_num)
        )

    prompts = collection.cmetadata.get('prompts', {})
    prompts = prompts if prompts else {}

    # Main LLM configuration using get_model_config
    qa_llm_conf = get_model_config(
        'qa_completion_llm',
        user_config=chat_params.config,
        db_metadata=collection.cmetadata
    )
    qa_llm_conf['callbacks'] = [] if answer_callbacks is None else answer_callbacks
    qa_llm_conf['streaming'] = chat_params.streaming
    chatllm = load_chatmodel(qa_llm_conf)

    # Create Retriever
    retriever = create_conversation_retriever(
        collection=collection,
        chat_params=chat_params,
        llm=chatllm
    )

    # Chain to rewrite question with history using get_model_config
    fup_llm_conf = get_model_config(
        'qa_followup_llm',
        user_config=chat_params.config,
        db_metadata=collection.cmetadata
    )
    fup_llm = load_chatmodel(fup_llm_conf)
    fup_chain = (
        load_condense_prompt(prompts.get('condense'))
        | fup_llm
        | StrOutputParser()
    )

    # Chain with "stuff document" type
    document_chain = create_stuff_documents_chain(
        llm=chatllm,
        prompt=load_qa_prompt(prompts)
    )

    retrieval_docs = (lambda x: x["question"]) | retriever
    retrivial_chain = (
        RunnablePassthrough.assign(
            context=retrieval_docs.with_config(run_name="retrieve_documents"),
        ).assign(answer=document_chain)
    ).with_config(run_name="retrieval_chain")

    # Final retrieval chain with proper input handling
    return (
        RunnablePassthrough.assign(
            question=fup_chain
        )
        | retrivial_chain
    )


def conversation_chain(
    chat_params: ChatParams,
    answer_callbacks: list[BaseCallbackHandler] | None = None,
) -> Chain:
    """
    Create a simple conversation chain for conversation tasks without a collection.
    This chain is used for general chat interactions that do not involve a specific
    collection or dataset.
    """

    # Chain to rewrite question with history
    # Check if followup_llm config is provided in chat_params
    fup_llm_conf = get_model_config(
        'qa_followup_llm',
        user_config=chat_params.config
    )

    fup_llm = load_chatmodel(fup_llm_conf)
    fup_chain = (
        load_condense_prompt()
        | fup_llm
        | StrOutputParser()
    )

    # Main LLM configuration
    # Check if completion_llm config is provided in chat_params
    llm_conf = get_model_config(
        'qa_completion_llm',
        user_config=chat_params.config
    )

    llm_conf['callbacks'] = [] if answer_callbacks is None else answer_callbacks
    llm_conf['streaming'] = chat_params.streaming

    prompt = PromptTemplate(
        input_variables=["question"],
        template="\n{question}",
    )
    llm = load_chatmodel(llm_conf)
    chain = prompt | llm

    return (
        RunnablePassthrough.assign(
            question=fup_chain
        )
        | chain
    )



"""Migration to JSONB format

Revision ID: 95d93297832c
Revises: b34017752e45
Create Date: 2024-05-10 18:25:42.596761

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import JSONB, JSON

# revision identifiers, used by Alembic.
revision = '95d93297832c'
down_revision = 'b34017752e45'
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.alter_column(
        'langchain_pg_embedding',
        'cmetadata',
        existing_type=sa.TEXT,
        type_=JSONB,
        postgresql_using='cmetadata::jsonb'
    )


def downgrade() -> None:
    op.alter_column(
        'langchain_pg_embedding',
        'cmetadata',
        existing_type=sa.TEXT,
        type_=JSON,
        postgresql_using='cmetadata::json'
    )



"""UUID utility functions"""
import uuid


def is_valid_uuid(val) -> bool:
    """ Check UUID validity """
    try:
        uuid.UUID(str(val))
        return True
    except (ValueError, TypeError):
        return False



"""Download files endpoint."""
import os
from fastapi import APIRouter, HTTPException, status
from fastapi.responses import FileResponse
from brevia.settings import get_settings


router = APIRouter()


@router.api_route(
    '/download/{file_path:path}',
    methods=['GET', 'HEAD'],
    tags=['Download'],
)
async def download_file(file_path: str):
    """
    Endpoint to download a file from the internal file system.
    """
    base_path = get_settings().file_output_base_path
    # Check if the base path is an S3 file path
    if base_path.startswith('s3://'):
        raise HTTPException(
            status.HTTP_404_NOT_FOUND,
            'File download is not supported',
        )

    full_path = os.path.join(base_path, file_path)
    if not os.path.isfile(full_path):
        raise HTTPException(
            status.HTTP_404_NOT_FOUND,
            f'File not found: {file_path}',
        )

    return FileResponse(full_path, filename=os.path.basename(full_path))



"""File output utilities"""
import tempfile
import os
from brevia.settings import get_settings


class LinkedFileOutput:
    """
    A class to handle file output operations in Brevia that need a download link.
    """
    job_id = None

    def __init__(self, job_id: str = None):
        """
        Initialize the FileOutput object with a job ID.

        :param job_id: The job ID to associate with this output.
        """
        self.job_id = job_id

    def file_path(self, filename: str):
        """
        Generate the file path for the output file.

        :param filename: The name of the file.
        :return: The full path to the file.
        """
        out_dir = get_settings().file_output_base_path
        if out_dir.startswith('s3://'):
            with tempfile.NamedTemporaryFile(prefix="brevia_", delete=False) as t_file:
                return t_file.name
        if self.job_id:
            out_dir = f"{out_dir}/{self.job_id}"
            os.makedirs(out_dir, exist_ok=True)

        return f"{out_dir}/{filename}"

    def file_url(self, filename: str):
        """
        Generate the URL for the output file.

        :param filename: The name of the file.
        :return: The URL to access the file.
        """
        # Generate the output URL
        base_url = get_settings().file_output_base_url
        return f'{base_url}/{filename}'

    def _s3_upload(self, file_path: str, bucket_name: str, object_name: str):
        """
        Upload the file to S3.

        :param file_path: The path to the file to upload.
        :param bucket_name: The name of the S3 bucket.
        :param object_name: The S3 object name.
        """
        try:
            import boto3  # pylint: disable=import-outside-toplevel
            s3 = boto3.client('s3')
            return s3.upload_file(file_path, bucket_name, object_name)
        except ModuleNotFoundError:
            raise ImportError('Boto3 is not installed!')

    def write(self, content: str, filename: str):
        """
        Write content of the file to the specified filename.
        Returns the URL of the file.

        :param content: The content to write to the file.
        :param filename: The name of the file to write to.
        """
        output_path = self.file_path(filename)
        with open(output_path, 'w', encoding='utf-8') as file:
            file.write(content)

        if self.job_id:
            filename = f"{self.job_id}/{filename}"
        base_path = get_settings().file_output_base_path
        if base_path.startswith('s3://'):
            # Extract bucket name and object name from S3 path
            bucket_name = base_path.split('/')[2]
            object_name = '/'.join(base_path.split('/')[3:]).lstrip('/')
            object_name += f"/{filename}"
            self._s3_upload(output_path, bucket_name, object_name.lstrip('/'))
            # Remove the local temp file
            os.unlink(output_path)

        return self.file_url(filename)



"""API endpoints definitions to show available providers and models"""
from fastapi import APIRouter, HTTPException, status
from brevia.dependencies import get_dependencies
from brevia.providers import list_providers, single_provider

router = APIRouter()


@router.api_route(
    '/providers',
    methods=['GET'],
    dependencies=get_dependencies(json_content_type=False),
    tags=['Providers'],
)
def api_providers(list_models: bool = True):
    """ /providers endpoint, list available providers and models """
    return list_providers(list_models=list_models)


@router.api_route(
    '/providers/{provider}',
    methods=['GET'],
    dependencies=get_dependencies(json_content_type=False),
    tags=['Providers'],
)
def provider_models(provider: str):
    """ /providers/{provider} endpoint, list available models for a provider """
    res = single_provider(provider=provider)
    if not res:
        raise HTTPException(
            status.HTTP_404_NOT_FOUND,
            f"Providers '{provider}' not found",
        )

    return res



"""Base retriever module"""
from langchain_core.documents import Document
from langchain_core.callbacks.manager import AsyncCallbackManagerForRetrieverRun
from langchain_core.vectorstores import VectorStore, VectorStoreRetriever


class BreviaBaseRetriever(VectorStoreRetriever):
    """ Base custom Retriever for BREVIA"""

    vectorstore: VectorStore
    """VectorStore used for retrieval."""

    search_kwargs: dict
    """Configuration containing settings for the search from the application"""

    async def _aget_relevant_documents(
        self, query: str, *, run_manager: AsyncCallbackManagerForRetrieverRun, **kwargs
    ) -> list[Document]:
        """
        Asynchronous implementation for retrieving relevant documents with score
        Merges results from multiple custom searches using different filters.

        Parameters:
            query (str): The search query.
            run_manager (CallbackManagerForRetrieverRun): Manager for retriever runs.

        Returns:
            List[Document]: A list of relevant documents based on the search.
        """
        if self.search_type == "similarity":
            docs = await self.vectorstore.asimilarity_search(
                query, **self.search_kwargs
            )
        elif self.search_type == "similarity_score_threshold":
            docs_and_similarities = (
                await self.vectorstore.asimilarity_search_with_relevance_scores(
                    query, **self.search_kwargs
                )
            )
            for doc, score in docs_and_similarities:
                doc.metadata["score"] = score
            docs = [doc for doc, _ in docs_and_similarities]
        elif self.search_type == "mmr":
            docs = await self.vectorstore.amax_marginal_relevance_search(
                query, **self.search_kwargs
            )
        else:
            msg = f"search_type of {self.search_type} not allowed."
            raise ValueError(msg)
        return docs



""" Callback class to handle conversation chain events """
from typing import Dict, Any, List, Sequence, Union
from uuid import UUID
import asyncio
import logging
import json
from glom import glom
from langchain_community.callbacks import OpenAICallbackHandler, get_openai_callback
from langchain_core.callbacks import AsyncCallbackHandler, BaseCallbackHandler
from langchain_core.documents import Document
from langchain_core.messages import BaseMessage
from langchain_core.outputs import LLMResult
from brevia.chat_history import add_history


# Only OpenAI token usage callback handler is supported for now
# other LLMs will always return 0 as tokens usage (for now)
token_usage_callback = get_openai_callback
TokensCallbackHandler = OpenAICallbackHandler


def token_usage(callb: TokensCallbackHandler) -> dict[str, int | float]:
    """Tokens usage and costs details (only OpenAI for now)"""
    return {
        'completion_tokens': callb.completion_tokens,
        'prompt_tokens': callb.prompt_tokens,
        'total_tokens': callb.total_tokens,
        'successful_requests': callb.successful_requests,
        'total_cost': callb.total_cost,
    }


class ConversationCallbackHandler(AsyncCallbackHandler):
    """Call back handler to return conversation chain results"""
    documents: List[Document] = []
    answer: str = ''
    chain_ended = asyncio.Event()

    async def on_chat_model_start(
        self,
        serialized: Dict[str, Any],
        messages: List[List[BaseMessage]],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> Any:
        """Run when a chat model starts running."""

    async def on_chain_start(
        self,
        serialized: Dict[str, Any],
        inputs: Dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> None:
        """Run when chain starts running."""
        if parent_run_id is None:
            self.chain_ended.clear()

    async def on_retriever_end(
        self,
        documents: Sequence[Document],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        tags: List[str] | None = None,
        **kwargs: Any,
    ) -> None:
        """Run on retriever end."""
        self.documents = documents

    async def on_chain_end(
        self,
        outputs: Dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> None:
        """Run when chain ends running."""
        if parent_run_id is None:
            self.answer = glom(
                outputs,
                'answer',
                default=glom(outputs, 'content', default='')
            )
            self.chain_ended.set()

    async def wait_conversation_done(self):
        """Wait for conversation end"""
        await self.chain_ended.wait()  # await until event would be .set()

    def chain_result(
        self,
        callb: TokensCallbackHandler,
        question: str,
        collection: str,
        x_chat_session: str | None = None,
    ) -> str:
        """Save chat history and add history id and source_documents to chain result"""
        chat_hist = add_history(
            session_id=x_chat_session,
            collection=collection,
            question=question,
            answer=self.answer.strip(" \n"),
            metadata=token_usage(callb),
        )

        docs = [{'chat_history_id': None if chat_hist is None else str(chat_hist.uuid)}]

        for doc in self.documents:
            docs.append({
                'page_content': doc.page_content,
                'metadata': doc.metadata
            })

        return json.dumps(docs)


class AsyncLoggingCallbackHandler(AsyncCallbackHandler):
    """Callback handler to handle logging in async calls"""
    log = None

    def __init__(self):
        self.log = logging.getLogger(__name__)
        super().__init__()

    async def on_chat_model_start(
        self,
        serialized: Dict[str, Any],
        messages: List[List[BaseMessage]],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> Any:
        """Run when a chat model starts running."""
        item = f'{run_id} / {parent_run_id}'
        self.log.info("%s - serialized: %s", item, serialized)
        self.log.info("%s - messages: %s", item, messages)

    async def on_chain_start(
        self,
        serialized: Dict[str, Any],
        inputs: Dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> None:
        """Run when chain starts running."""
        item = f'{run_id} / {parent_run_id}'
        self.log.info("%s - serialized: %s", item, serialized)
        self.log.info("%s - inputs: %s", item, inputs)

    async def on_chain_end(
        self,
        outputs: Dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> None:
        """Run when chain ends running."""
        item = f'{run_id} / {parent_run_id}'
        self.log.info("%s - outputs: %s", item, outputs)

    async def on_retriever_start(
        self,
        serialized: Dict[str, Any],
        query: str,
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        tags: List[str] | None = None,
        metadata: Dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        """Run on retriever start."""
        item = f'{run_id} / {parent_run_id}'
        self.log.info("%s - query: %s", item, query)
        self.log.info("%s - serialized: %s", item, serialized)

    async def on_retriever_end(
        self,
        documents: Sequence[Document],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        tags: List[str] | None = None,
        **kwargs: Any,
    ) -> None:
        """Run on retriever end."""
        item = f'{run_id} / {parent_run_id}'
        self.log.info("%s - documents: %s", item, documents)

    async def on_retriever_error(
        self,
        error: Union[Exception, KeyboardInterrupt],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        tags: List[str] | None = None,
        **kwargs: Any,
    ) -> None:
        """Run on retriever error."""
        item = f'{run_id} / {parent_run_id}'
        self.log.error("%s - %s", item, error)

    # async def on_text(
    #     self,
    #     text: str,
    #     *,
    #     run_id: UUID,
    #     parent_run_id: UUID | None = None,
    #     **kwargs: Any,
    # ) -> Any:
    #     """Run on arbitrary text."""
        # self.log.info('on_text')
        # self.log.info(text)


class LoggingCallbackHandler(BaseCallbackHandler):
    """Callback handler to handle logging in async calls"""
    log = None

    def __init__(self):
        self.log = logging.getLogger(__name__)
        super().__init__()

    def on_chat_model_start(
        self,
        serialized: Dict[str, Any],
        messages: List[List[BaseMessage]],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> Any:
        """Run when a chat model starts running."""
        item = f'{run_id} / {parent_run_id}'
        self.log.info("%s - serialized: %s", item, serialized)
        self.log.info("%s - messages: %s", item, messages)

    def on_chain_start(
        self,
        serialized: Dict[str, Any],
        inputs: Dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> None:
        """Run when chain starts running."""
        item = f'{run_id} / {parent_run_id}'
        self.log.info("%s - serialized: %s", item, serialized)
        self.log.info("%s - inputs: %s", item, inputs)

    def on_chain_end(
        self,
        outputs: Dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> None:
        """Run when chain ends running."""
        item = f'{run_id} / {parent_run_id}'
        self.log.info("%s - outputs: %s", item, outputs)

    def on_llm_start(
        self,
        serialized: Dict[str, Any],
        prompts: List[str],
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        tags: List[str] | None = None,
        metadata: Dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> Any:
        """Run when LLM starts running."""
        item = f'{run_id} / {parent_run_id}'
        self.log.info("%s - serialized: %s", item, serialized)
        self.log.info("%s - prompts: %s", item, prompts)

    def on_llm_end(
        self,
        response: LLMResult,
        *,
        run_id: UUID,
        parent_run_id: UUID | None = None,
        **kwargs: Any,
    ) -> Any:
        """Run when LLM ends running."""
        item = f'{run_id} / {parent_run_id}'
        self.log.info("%s - response: %s", item, response)

    # def on_text(
    #     self,
    #     text: str,
    #     *,
    #     run_id: UUID,
    #     parent_run_id: UUID | None = None,
    #     **kwargs: Any,
    # ) -> Any:
    #     """Run on arbitrary text."""
    #     self.log.info('on_text')
    #     self.log.info(text)



"""API endpoints definitions to handle audio input"""
import logging
from typing import Annotated
from fastapi import APIRouter, UploadFile, Form
from brevia.models import load_audiotranscriber
from brevia.dependencies import (
    get_dependencies,
    save_upload_file_tmp,
)

router = APIRouter()


@router.post(
    '/transcribe',
    dependencies=get_dependencies(json_content_type=False),
    tags=['Analysis'],
)
def audio_transcriptions(
    file: UploadFile,
    language: Annotated[str, Form()],
):
    """ /transcribe endpoint, audio file transcription """
    log = logging.getLogger(__name__)
    log.info("Uploaded '%s' - %s - %s", file.filename, file.content_type, file.size)
    log.info("Language '%s'", language)
    tmp_file_path = save_upload_file_tmp(file)

    audio = load_audiotranscriber()
    result = audio.transcribe(
        file=tmp_file_path,
        language=language,
    )
    log.info('Audio transcription completed')
    log.info(result)
    return result



"""API endpoints definitions to handle document analysis like summarization"""
from typing import Annotated
from base64 import b64decode
from pathlib import Path
import tempfile
import json
import logging
from pydantic import BaseModel
from fastapi import (
    APIRouter,
    BackgroundTasks,
    HTTPException,
    status,
    UploadFile,
    Form,
)
from brevia.dependencies import get_dependencies, save_upload_file_tmp
from brevia import async_jobs, load_file
from brevia.services import SummarizeTextService

router = APIRouter()


class SummarizeBody(BaseModel):
    """ summarize input """
    text: str
    chain_type: str | None = None
    initial_prompt: dict | None = None
    iteration_prompt: dict | None = None
    token_data: bool = False


@router.post('/summarize', dependencies=get_dependencies(), tags=['Analysis'])
def summarize_text(summarize: SummarizeBody):
    """
    /summarize endpoint:
    Summarize text and return a summary with algorithm selection
    and custom prompt options.

    Args:
        summarize (SummarizeBody): An object representing the request data.
            It contains the following parameters:
            - text (str): The text to be summarized.
            - chain_type (str, optional): The main Langchain summarization chain type.
                Should be one of "stuff", "map_reduce", or "refine".
                If not provided, "stuff" is used by default.
            - initial_prompt (dict, optional): Optional custom prompt as a JSON object
                to be used in the selected Langchain chain type to replace
                the main chain prompt defaults.
            - iteration_prompt (dict, optional): Optional custom prompts to be used
                in the selected algorithm. This prompt will serve as the iterative
                prompt for various summarization algorithms. For instance, if the user
                selects the "map_reduce" algorithm, the iteration prompt is used
                as the "combine_prompt".
            - token_data (bool): A boolean indicating whether to include
                token-level data in the summary.

    Returns:
        dict: A JSON object representing the summary of the provided text.

    Raises:
        HTTPException: If the 'text' field is empty, with status code 400.

    This endpoint allows you to choose the summarization algorithm from those available
    in the Langchain library and customize the prompts based on the chosen algorithm.
    """
    summarize.text = load_file.cleanup_text(summarize.text).strip()
    if not summarize.text:
        raise HTTPException(
            status.HTTP_400_BAD_REQUEST,
            'Empty text field',
        )

    service = SummarizeTextService()
    return service.run(summarize.model_dump())


@router.post(
    '/upload_summarize',
    dependencies=get_dependencies(json_content_type=False),
    tags=['Analysis'],
)
def upload_summarize(
    background_tasks: BackgroundTasks,
    chain_type: Annotated[str, Form()] = '',
    initial_prompt: Annotated[str, Form()] = '',
    iteration_prompt: Annotated[str, Form()] = '',
    file: UploadFile | None = None,
    file_content: Annotated[str, Form()] = '',
    token_data: Annotated[bool, Form()] = False,
    payload: Annotated[str, Form()] = '{}',
):
    """
    Upload a PDF file and perform summarization with algorithm
    selection and custom prompt options.

    Args:
        - background_tasks (BackgroundTasks): A background task manager
            for handling asynchronous tasks
        - chain_type (str, optional): The main Langchain summarization chain type
            Should be one of "stuff", "map_reduce", or "refine"
            If not provided, 'stuff' is used by default
        - initial_prompt (dict, optional): Optional custom prompt as a JSON object
                to be used in the selected Langchain chain type to replace
                the main chain prompt defaults.
        - iteration_prompt (dict, optional): Optional custom prompts to be used
            in the selected algorithm. This prompt will serve as the iterative
            prompt for various summarization algorithms. For instance, if the user
            selects the "map_reduce" algorithm, the iteration prompt is used
            as the "combine_prompt".
        - file (UploadFile | None): An uploaded PDF file to be summarized
        - file_content (str, optional): Content of the PDF file provided
            as a base64-encoded string
        - token_data (bool, optional): A boolean indicating whether to include
            token-level data in the summary
        - payload (str, optional): Optional payload in JSON format to use
            in the async job service to add custom options and custom fields

    Returns:
        A JSON object representing the job UUID for the asynchronous summarization task

    Raises:
        HTTPException with status code 400 if either 'file' or 'file_content'
        form field is missing

    With this endpoint, you can upload a PDF file for summarization, select a
    summarization algorithm from those available in the Langchain library,
    and customize the prompts according to your chosen algorithm
    """
    if file:
        log = logging.getLogger(__name__)
        log.info("Uploaded '%s' - %s - %s", file.filename, file.content_type, file.size)
        log.info(
            "Type '%s' - initial prompt: '%s' - iteration prompt: '%s'",
            chain_type,
            initial_prompt,
            iteration_prompt
        )
        tmp_path = save_upload_file_tmp(file)
    elif file_content:
        tmp_path = save_base64_tmp_file(file_content=file_content)
    else:
        raise HTTPException(
            status.HTTP_400_BAD_REQUEST,
            'One of "file" or "file_content" form field is mandatory',
        )

    job = async_jobs.create_job(
        service='brevia.services.SummarizeFileService',
        payload=json.loads(payload) | {
            'file_path': tmp_path,
            'chain_type': chain_type,
            'initial_prompt':
                json.loads(initial_prompt) if initial_prompt else None,
            'iteration_prompt':
                json.loads(iteration_prompt) if iteration_prompt else None,
            'token_data': token_data,
        }
    )
    background_tasks.add_task(async_jobs.run_job_service, job.uuid)

    return {'job': job.uuid}


@router.post(
    '/upload_analyze',
    dependencies=get_dependencies(json_content_type=False),
    tags=['Analysis'],
)
def upload_analyze(
    file: UploadFile,
    service: Annotated[str, Form()],
    background_tasks: BackgroundTasks,
    payload: Annotated[str, Form()] = '{}',
):
    """
    Upload a file and perform some analysis using a `service` class
    """
    log = logging.getLogger(__name__)
    log.info("Uploaded '%s' - %s - %s", file.filename, file.content_type, file.size)
    tmp_path = save_upload_file_tmp(file)

    payload = json.loads(payload)
    payload['file_path'] = tmp_path
    payload.setdefault('file_name', file.filename)
    job = async_jobs.create_job(
        service=service,
        payload=payload,
    )
    background_tasks.add_task(async_jobs.run_job_service, job.uuid)

    return {'job': job.uuid}


def save_base64_tmp_file(file_content: str) -> str:
    """ Save base64 file content to temp file, return path """

    with tempfile.NamedTemporaryFile(delete=False) as tmp:
        content_as_bytes = str.encode(file_content)  # convert string to bytes
        content_recovered = b64decode(content_as_bytes)  # decode base64string
        with open(tmp.name, 'wb') as file:
            file.write(content_recovered)

    return str(Path(tmp.name))



""" Functions to handle prompts."""

from langchain_core.prompts import (
    ChatPromptTemplate,
    HumanMessagePromptTemplate,
    PromptTemplate,
    SystemMessagePromptTemplate,
)
from langchain_core.prompts.base import BasePromptTemplate
from langchain_core.prompts.loading import (
    load_prompt_from_config,
    _load_few_shot_prompt,
)
import yaml
from brevia.settings import get_settings


def prompt_file_path(relative_path: str):
    """ Get the full path of a prompt file, given the relative path """
    return f'{get_settings().prompts_base_path}/{relative_path}'


def load_prompt_from_yaml(relative_path: str) -> PromptTemplate:
    """ Load a prompt from a yaml file, given the prompt file relative path """
    file_path = prompt_file_path(relative_path)
    with open(file_path) as f:
        config = yaml.safe_load(f)
    if not config:
        raise ValueError(f'Failed to load prompt from {file_path}')
    type = config.pop('_type', 'prompt')
    if type == 'few_shot':
        return _load_few_shot_prompt(config)

    return PromptTemplate(**config)


def load_qa_prompt(prompts: dict | None = None) -> ChatPromptTemplate:
    """
        Load prompt for RAG-Q/A, following system/human/ai pattern.
        If prompts are provided, they will overwrite the default prompts.
        Otherwise, the default prompts will be loaded from the default path:
            - rag/human.yml: human prompt, the user input
            - rag/system.yml: system prompt, base instruction for the model behavior
    """
    human = load_prompt_from_yaml('rag/human.yml')
    system = load_prompt_from_yaml('rag/system.yml')

    if prompts:
        if prompts.get('system'):
            system = load_prompt_from_config(prompts.get('system'))
        if prompts.get('human'):
            human = load_prompt_from_config(prompts.get('human'))

    system_message_prompt = SystemMessagePromptTemplate(prompt=system)
    human_message_prompt = HumanMessagePromptTemplate(prompt=human)

    return ChatPromptTemplate.from_messages(
        [system_message_prompt, human_message_prompt]
    )


def load_condense_prompt(config: dict | None = None) -> BasePromptTemplate:
    """
        Load prompt for RAG-Condense, to create the follow-up question
        in a conversation.
        Prompt config is loaded if provided, otherwise, it will load
        the default 'rag/condense.yml' prompt.
    """
    if config:
        return load_prompt_from_config(config=config)

    return load_prompt_from_yaml('rag/condense.yml')


def load_stuff_prompts(prompts: dict | None) -> dict:
    """Load custom prompts for the 'stuff' summarization chain.
    Summarization chain is the simplest one and have only one prompt

    Args:
        prompts (dict | None): Optional prompts for customization.
        It should be a dictionary with the following keys:
        - 'initial_prompt' (dict): The prompt to use for the initial summarization.

    Returns:
        dict: Custom prompts for 'stuff' summarization chain.
    """
    if prompts.get('initial_prompt'):
        return {
            'prompt': load_prompt_from_config(prompts.get('initial_prompt'))
        }

    return {}


def load_map_prompts(prompts: dict | None) -> dict:
    """Load custom prompts for the 'map_reduce' summarization chain.
    The 'map_reduce' summarization chain utilizes a more sophisticated algorithm that
    merges multiple prompts to create a summary. It divides the text into text chunks
    and applies the initial prompt to each of them. Afterward, it uses the iteration
    prompt to combine all the result into a final summary.

    Args:
        prompts (dict | None): Optional prompts for customization.
        It should be a dictionary with the following keys:
        - 'initial_prompt' (dict): The prompt to use for the chunks summarization.
        - 'iteration_prompt' (dict | None): (Optional) The prompt to use for summarizing
            all the pieces.

    Returns:
        dict: Custom prompts for 'map_reduce' summarization chain.
    """
    prompts_data = {}
    if prompts.get('initial_prompt'):
        map_prompt = load_prompt_from_config(prompts.get('initial_prompt'))
        prompts_data['map_prompt'] = map_prompt
        prompts_data['combine_prompt'] = map_prompt
        if prompts.get('iteration_prompt'):
            combine_prompt = load_prompt_from_config(prompts.get('iteration_prompt'))
            prompts_data['combine_prompt'] = combine_prompt

    return prompts_data


def load_refine_prompts(prompts: dict | None) -> dict:
    """Load custom prompts for the 'refine' summarization chain.

    The 'refine' summarization chain breaks the document into chunk pieces, then
    concentrates on making the summary better by refining it.
    It starts using the initial prompt for the first chunk and then uses the
    iteration prompt over the remaining chunks to further enhance the summary.

    In LangChain, the default prompt names for the refine algorithm are:
    - question_prompt: The prompt used for summarizing the first chunk.
    - refine_prompt: The prompt used for refining the summary over the remaining chunks.

    Args:
        prompts (dict | None): Optional prompts for customization.
        It should be a dictionary with the following keys:
        - 'initial_prompt' (dict): The prompt to use for the first chunk summarization.
        - 'iteration_prompt' (dict | None): (Optional) The prompt to use for increasing
          information over the chunk pieces.

    Returns:
        dict: Custom prompts for the 'refine' summarization chain.
    """

    prompts_data = {}
    if prompts.get('initial_prompt'):
        question_prompt = load_prompt_from_config(prompts.get('initial_prompt'))
        prompts_data['question_prompt'] = question_prompt
        if prompts.get('iteration_prompt'):
            refine_prompt = load_prompt_from_config(prompts.get('iteration_prompt'))
            prompts_data['refine_prompt'] = refine_prompt

    return prompts_data



"""Read text from a file with different formats"""
import os
import re
from importlib import import_module
import mimetypes
import tempfile
from typing import List, Any
from urllib.parse import urlparse
from pathlib import Path
import requests
from bs4 import BeautifulSoup
from langchain_community.document_loaders.html_bs import BSHTMLLoader
from langchain_community.document_loaders.csv_loader import CSVLoader
from langchain_community.document_loaders.pdf import PyPDFLoader, UnstructuredPDFLoader
from langchain_community.document_loaders.text import TextLoader
from langchain_core.documents import Document


def cleanup_text(text_in: str) -> str:
    """ Cleanup input text from multple spaces and newlines"""
    new_text = re.sub(r'\n\n+', '\n\n', text_in)
    new_text = new_text.replace("\x00", " ")
    new_text = re.sub(r'\.\.\.+', '...', new_text)

    return " ".join(new_text.strip().split(' '))


def read_pdf_file(
    file_path: str,
    page_from: int = None,  # from page, used in PDF
    page_to: int = None,  # to page, used in PDF
    **loader_kwargs: Any,
) -> str:
    """
    Load PDF file and return its textual content
    """
    docs = load_documents_pdf(file_path=file_path, **loader_kwargs)
    if page_from is not None:
        docs = [x for x in docs if x.metadata['page'] >= (page_from - 1)]
    if page_to is not None:
        docs = [x for x in docs if x.metadata['page'] <= (page_to - 1)]

    return ' '.join([cleanup_text(item.page_content) for item in docs]).strip()


def load_documents_pdf(file_path: str, **kwargs: Any) -> List[Document]:
    """
    Load documents from PDF file, first try to extract text with PyPDF,
    then use Unstructured and OCR to read
    """
    loader = PyPDFLoader(file_path=file_path)
    docs = loader.load()

    # check if `ocr` argument is passed and not falsy
    if 'ocr' not in kwargs or not kwargs['ocr']:
        return docs

    # try with OCR if extracted text is too short - optional 'min_text_len' parameter
    min_text_len = kwargs['min_text_len'] if 'min_text_len' in kwargs else 100
    text = ' '.join([cleanup_text(item.page_content) for item in docs]).strip()
    if len(text) > min_text_len:
        return docs

    ocr_languages = kwargs['ocr_languages'] if 'ocr_languages' in kwargs else None
    loader = UnstructuredPDFLoader(
        file_path=file_path, mode='elements', strategy='ocr_only',
        ocr_languages=ocr_languages,
    )

    return loader.load()


def read_txt_file(
    file_path: str,
) -> str:
    """
    Load TXT file and return its content
    """
    with open(file_path, 'r') as file:  # pylint: disable=unspecified-encoding
        text = file.read()

    return cleanup_text(text).strip()


def read_csv_file(
    file_path: str,
) -> str:
    """
    Load CSV file and return its content
    """
    docs = load_documents_csv(file_path=file_path)

    return ' '.join([cleanup_text(item.page_content) for item in docs]).strip()


def read(
    file_path: str,
    **loader_kwargs: Any,
) -> str:
    """
    Load text from a file, see `MIME_TYPES_READERS` for supported content types
    """

    if not os.path.isfile(file_path):
        raise FileNotFoundError(file_path)

    mtype = mimetypes.guess_type(file_path)[0]
    if mtype not in MIME_TYPES_LOADERS:
        raise ValueError(f'Unsupported file content type "{mtype}"')

    read_function = MIME_TYPES_READERS[mtype]

    return read_function(file_path, **loader_kwargs)


def read_html_url(
    url: str,
    **loader_kwargs: Any,
) -> str:
    """
    Load text from HTML content of web page URL
    """
    parsed_url = urlparse(url)
    remote = True if parsed_url.scheme not in ['file', ''] else False
    temp_file = tempfile.NamedTemporaryFile(delete=False)
    if remote:
        response = requests.get(url)
        response.raise_for_status()
        text = response.text
    else:
        text = Path(url).read_text()

    with open(temp_file.name, 'w') as file:  # pylint: disable=missing-timeout
        content = filter_html(
            html=text, selector=loader_kwargs.pop('selector', ''),
            callback=loader_kwargs.pop('callback', ''))
        file.write(content)
    loader = BSHTMLLoader(file_path=temp_file.name, **loader_kwargs)
    docs = loader.load()
    os.unlink(temp_file.name)

    return ' '.join([cleanup_text(item.page_content) for item in docs]).strip()


def filter_html(html: str, selector: str = '', callback: str = '') -> str:
    """Filter HTML content using callbacks and selector rule, using BeautifulSoup """
    if callback:
        module_name, func_name = callback.rsplit('.', 1)
        try:
            module = import_module(module_name)
            html = getattr(module, func_name)(html)  # pylint: disable=not-callable
        except (ImportError, AttributeError):
            raise ValueError(f'Callback "{callback}" not found')
    if not selector:
        return html
    soup = BeautifulSoup(html, features='lxml')
    tags = soup.select(selector=selector)

    return ' '.join([tag.decode_contents() for tag in tags])


def load_documents(file_path: str, **kwargs: Any) -> List[Document]:
    """Load documents from a file or folder path"""
    mtype = mimetypes.guess_type(file_path)[0]
    if mtype not in MIME_TYPES_LOADERS:
        raise ValueError(f'Unsupported file content type "{mtype}"')
    load_function = MIME_TYPES_LOADERS[mtype]

    return load_function(file_path, **kwargs)


def load_documents_txt(file_path: str, **kwargs: Any) -> List[Document]:
    """Load documents from TXT file"""
    loader = TextLoader(file_path=file_path, autodetect_encoding=True)
    return loader.load()


def load_documents_html(file_path: str, **kwargs: Any) -> List[Document]:
    """Load documents from HTML file"""
    loader = BSHTMLLoader(file_path=file_path, **kwargs)
    return loader.load()


def load_documents_csv(file_path: str, **kwargs: Any) -> List[Document]:
    """Load documents from CSV file"""
    loader = CSVLoader(
        file_path=file_path,
        csv_args={
            'delimiter': ',',
            'quotechar': '"',
            'doublequote': True,
            'skipinitialspace': False,
        }
    )
    return loader.load()


MIME_TYPES_LOADERS = {
    'application/pdf': load_documents_pdf,
    'text/plain': load_documents_txt,
    'text/csv': load_documents_csv,
    'text/html': load_documents_html,
}

MIME_TYPES_READERS = {
    'application/pdf': read_pdf_file,
    'text/plain': read_txt_file,
    'text/csv': read_csv_file,
    'text/html': read_html_url,
}



"""Base class for analysis tasks"""
from abc import ABC, abstractmethod
from typing import Any


class BaseAnalysisTask(ABC):
    """Base class for analysis tasks"""
    prompts: dict[str, Any] = {}

    @abstractmethod
    def perform_task(self) -> dict:
        """Perform task logic using payload"""

    @abstractmethod
    def load_analysis_prompts(self, prompts: dict | None = None):
        """Load analysis prompts usign optional input prompts"""



""" Module to list available providers and models """
import json
import logging
from os import environ
from openai import OpenAI, OpenAIError
from cohere import Client as CohereClient
from ollama import Client as OllamaClient
from anthropic import Client as AnthropicClient
from brevia.connection import db_connection
from brevia.settings import get_settings, update_db_conf


def list_providers(list_models: bool = True) -> list:
    """ List available providers and models """
    providers = []
    for provider in PROVIDER_MODELS_MAP.keys():
        item = {'model_provider': provider}
        if list_models:
            models = PROVIDER_MODELS_MAP.get(provider)()
            item['models'] = models
        providers.append(item)

    return providers


def single_provider(provider: str) -> dict | None:
    """ List available models for a provider """
    if provider not in PROVIDER_MODELS_MAP:
        return None

    models = PROVIDER_MODELS_MAP.get(provider)()
    return {'model_provider': provider, 'models': models}


def update_providers(force: bool = False):
    """ Update providers list in settings from API and save providers to DB """
    log = logging.getLogger(__name__)
    settings = get_settings()
    if settings.providers and not force:
        return
    try:
        providers = list_providers()
        log.info('Adding "providers" to brevia configuration DB')
        update_db_conf(db_connection(), {'providers': json.dumps(providers)})
    except Exception as exc:  # pylint: disable=broad-exception-caught
        log.error('Failed to update providers: %s', exc)


def load_openai_models() -> list | None:
    """ Load OpenAI models """
    try:
        client = OpenAI()
        list_models = client.models.list().model_dump()
    except OpenAIError:
        return None
    models = []
    for item in list_models['data']:
        name = item['id']
        if name.startswith('gpt-') or name.startswith('o1') or name.startswith('o3'):
            models.append({'name': name})

    return models


def load_deepseek_models() -> list | None:
    """ Load DeepSeek models """
    try:
        client = OpenAI(
            base_url='https://api.deepseek.com/v1',
            api_key=environ.get('DEEPSEEK_API_KEY'),
        )
        list_models = client.models.list().model_dump()
    except OpenAIError:
        return None
    models = []
    for item in list_models['data']:
        name = item['id']
        models.append({'name': name})

    return models


def load_anthropic_models() -> list | None:
    """ Load Anthropic models """
    try:
        client = AnthropicClient()
        list_models = client.models.list().model_dump()
    except Exception:  # pylint: disable=broad-exception-caught
        return None
    models = []
    for item in list_models['data']:
        name = item['id']
        models.append({'name': name})

    return models


def load_cohere_models():
    """ Load Cohere models """
    try:
        co = CohereClient()
        list_models = co.models.list(page_size=100).model_dump()
    except Exception:  # pylint: disable=broad-exception-caught
        return None
    models = []
    for item in list_models['models']:
        name = item['name']
        if 'chat' in item['endpoints']:
            models.append({'name': name, 'tokens_limit': int(item['context_length'])})

    return models


def load_ollama_models():
    """ Load Ollama models """
    try:
        ol = OllamaClient()
        list_models = ol.list()
    except Exception:  # pylint: disable=broad-exception-caught
        return None
    models = []
    for item in list_models['models']:
        name = item['model']
        fam = item['details']['family']
        if fam not in ['bert', 'nomic-bert']:
            models.append({'name': name})

    return models


PROVIDER_MODELS_MAP = {
    'openai': load_openai_models,
    'cohere': load_cohere_models,
    'anthropic': load_anthropic_models,
    'ollama': load_ollama_models,
    'deepseek': load_deepseek_models,
}



"""API endpoints for question answering and search"""
import asyncio
from typing import Annotated
from typing_extensions import Self
from glom import glom
from pydantic import Field, model_validator
from langchain.callbacks import AsyncIteratorCallbackHandler
from langchain.chains.base import Chain
from langchain_core.callbacks import BaseCallbackHandler
from langchain_core.messages.ai import AIMessage
from fastapi import APIRouter, Header
from fastapi.responses import StreamingResponse
from brevia import chat_history
from brevia.dependencies import (
    get_dependencies,
    check_collection_name,
)
from brevia.callback import (
    ConversationCallbackHandler,
    token_usage_callback,
    token_usage,
    TokensCallbackHandler,
)
from brevia.query import (
    SearchQuery,
    ChatParams,
    conversation_chain,
    conversation_rag_chain,
    search_vector_qa,
)
from brevia.models import test_models_in_use

router = APIRouter()


class ChatBody(ChatParams):
    """ /chat request body """
    question: str
    collection: str | None = None
    chat_history: list = []
    chat_lang: str | None = None
    config: dict | None = None
    mode: str = Field(pattern='^(rag|conversation)$', default='rag')
    token_data: bool = False

    @model_validator(mode='after')
    def check_collection(self) -> Self:
        """Validate collection and mode"""
        if not self.collection and self.mode == 'rag':
            raise ValueError('Collection required for rag mode')
        return self


@router.post('/prompt', dependencies=get_dependencies(), deprecated=True, tags=['Chat'])
@router.post('/chat', dependencies=get_dependencies(), tags=['Chat'])
async def chat_action(
    chat_body: ChatBody,
    x_chat_session: Annotated[str | None, Header()] = None,
):
    """
    /chat endpoint, ask chatbot about a collection of documents to perform a rag chat.
    If collection is not provided, it will use a simple completion chain.
    """
    # Check if collection is provided and valid
    collection = None
    if chat_body.collection:
        collection = check_collection_name(chat_body.collection)
        if not collection.cmetadata:
            collection.cmetadata = {}

    lang = chat_language(
        chat_body=chat_body,
        cmetadata=collection.cmetadata if collection else {}
    )
    conversation_handler = ConversationCallbackHandler()
    stream_handler = AsyncIteratorCallbackHandler()

    # Select chain based on chat_body.mode
    if chat_body.mode == 'rag':
        # RAG-based conversation chain using collection context
        chain = conversation_rag_chain(
            collection=collection,
            chat_params=ChatParams(**chat_body.model_dump()),
            answer_callbacks=[stream_handler] if chat_body.streaming else [],
        )
        embeddings = collection.cmetadata.get('embeddings', None)
    elif chat_body.mode == 'conversation':
        # Test mode - currently same as simple conversation
        chain = conversation_chain(
            chat_params=ChatParams(**chat_body.model_dump()),
            answer_callbacks=[stream_handler] if chat_body.streaming else [],
        )
        embeddings = None

    with token_usage_callback() as token_callback:
        if not chat_body.streaming or test_models_in_use():
            return await run_chain(
                chain=chain,
                chat_body=chat_body,
                lang=lang,
                token_callback=token_callback,
                x_chat_session=x_chat_session,
                embeddings=embeddings,
                chain_callbacks=[conversation_handler],
            )

        asyncio.create_task(run_chain(
            chain=chain,
            chat_body=chat_body,
            lang=lang,
            token_callback=token_callback,
            x_chat_session=x_chat_session,
            embeddings=embeddings,
            chain_callbacks=[conversation_handler],
        ))

        async def event_generator(
            stream_callback: AsyncIteratorCallbackHandler,
            conversation_callback: ConversationCallbackHandler,
            token_callback: TokensCallbackHandler,
            chat_body: ChatBody,
            x_chat_session: str | None = None,
        ):
            ait = stream_callback.aiter()

            async for token in ait:
                yield token

            await conversation_callback.wait_conversation_done()

            yield conversation_callback.chain_result(
                callb=token_callback,
                question=chat_body.question,
                collection=chat_body.collection,
                x_chat_session=x_chat_session,
            )

        return StreamingResponse(event_generator(
            stream_callback=stream_handler,
            conversation_callback=conversation_handler,
            token_callback=token_callback,
            chat_body=chat_body,
            x_chat_session=x_chat_session,
        ))


def chat_language(chat_body: ChatBody, cmetadata: dict) -> str:
    """Retrieve the language to be used in Q/A response"""
    chat_lang = chat_body.chat_lang or cmetadata.get('chat_lang')

    return chat_lang if chat_lang else ''


def retrieve_chat_history(history: list, question: str,
                          session: str = None, embeddings: dict | None = None) -> list:
    """Retrieve chat history to be used in final prompt creation"""
    chat_hist = chat_history.history(
        chat_history=history,
        session=session,
    )
    if chat_hist and not chat_history.is_related(chat_hist, question, embeddings):
        chat_hist = []

    return chat_hist


async def run_chain(
    chain: Chain,
    chat_body: ChatBody,
    lang: str,
    token_callback: TokensCallbackHandler,
    x_chat_session: str,
    embeddings: dict | None = None,
    chain_callbacks: list[BaseCallbackHandler] | None = None,
):
    """Run chain usign async methods and return result"""
    result = await chain.ainvoke({
        'question': chat_body.question,
        'chat_history': retrieve_chat_history(
            history=chat_body.chat_history,
            question=chat_body.question,
            session=x_chat_session,
            embeddings=embeddings,
        ),
        'lang': lang,
    },
        config={'callbacks': chain_callbacks},
        return_only_outputs=True,
    )
    return chat_result(
        result=result,
        callb=token_callback,
        chat_body=chat_body,
        x_chat_session=x_chat_session
    )


def chat_result(
    result: dict | AIMessage,
    callb: TokensCallbackHandler,
    chat_body: ChatBody,
    x_chat_session: str | None = None,
) -> dict:
    """ Handle chat result: save chat history and return answer """
    answer = glom(result, 'answer', default=glom(result, 'content', default=''))
    answer = str(answer).strip(" \n")

    chat_history_id = None
    if not chat_body.streaming:
        chat_hist = chat_history.add_history(
            session_id=x_chat_session,
            collection=chat_body.collection,
            question=chat_body.question,
            answer=answer,
            metadata=token_usage(callb),
        )
        chat_history_id = None if chat_hist is None else str(chat_hist.uuid)

    context = result['context'] if 'context' in result else None

    return {
        'bot': answer,
        'docs': None if not chat_body.source_docs else context,
        'chat_history_id': chat_history_id,
        'token_data': None if not chat_body.token_data else token_usage(callb)
    }


@router.post('/search', dependencies=get_dependencies(), tags=['Index'])
def search_documents(search: SearchQuery):
    """
        /search endpoint:
        Search the first {docs_num} relevant documents for a question

        Specify distance_strategy:
            EUCLIDEAN = EmbeddingStore.embedding.l2_distance (default)
            COSINE = EmbeddingStore.embedding.cosine_distance
            MAX_INNER_PRODUCT = EmbeddingStore.embedding.max_inner_product
    """
    collection = check_collection_name(search.collection)
    if search.docs_num is None and 'docs_num' in collection.cmetadata:
        search.docs_num = int(collection.cmetadata['docs_num'])
    result = search_vector_qa(search=search)

    return extract_content_score(result)


def extract_content_score(data_list) -> list:
    """Extract score from search result list."""

    if 'error' in data_list:
        return {'error': data_list['error']}
    extracted_data = []
    for item in data_list:
        content = item[0].page_content.strip(" \n")
        score = item[1]
        extracted_data.append({"content": content, "score": score})
    return extracted_data



"""Async Jobs table & utilities"""
import logging
import time
from datetime import datetime
from sqlalchemy import BinaryExpression, Column, desc, func, String, text
from pydantic import BaseModel as PydanticModel
from sqlalchemy.dialects.postgresql import JSON, TIMESTAMP, SMALLINT
from sqlalchemy.orm import Session, Query
from langchain_community.vectorstores.pgembedding import BaseModel
from brevia.connection import db_connection
from brevia.services import BaseService
from brevia.utilities.dates import date_filter
from brevia.utilities.json_api import query_data_pagination
from brevia.utilities.types import load_type

MAX_DURATION = 120  # default max duration is 120 min / 2hr
MAX_ATTEMPTS = 1  # default max number of attempts


class AsyncJobsStore(BaseModel):
    # pylint: disable=too-few-public-methods,not-callable
    """ Async Jobs table """
    __tablename__ = "async_jobs"

    service = Column(String(), nullable=False)
    payload = Column(JSON())
    expires = Column(TIMESTAMP(timezone=False))
    created = Column(
        TIMESTAMP(timezone=False),
        nullable=False,
        server_default=func.current_timestamp(),
    )
    completed = Column(TIMESTAMP(timezone=False))
    locked_until = Column(TIMESTAMP(timezone=False))
    max_attempts = Column(
        SMALLINT(),
        nullable=False,
        server_default='1',
    )
    result = Column(JSON(), nullable=True)


def single_job(uuid: str) -> (AsyncJobsStore | None):
    """ Get single job by UUID """
    with Session(db_connection()) as session:
        return session.get(AsyncJobsStore, uuid)


class JobsFilter(PydanticModel):
    """ Jobs filter """
    min_date: str | None = None
    max_date: str | None = None
    service: str | None = None
    completed: bool | None = None
    page: int = 1
    page_size: int = 50


def get_jobs(filter: JobsFilter) -> dict:  # pylint: disable=redefined-builtin
    """
    Read async jobs with optional filters using pagination data in response.
    """

    # Handle date filters - only apply if explicitly provided
    filter_min_date = text('1 = 1')  # always true by default
    filter_max_date = text('1 = 1')  # always true by default

    if filter.min_date:
        min_date = date_filter(filter.min_date, 'min')
        filter_min_date = AsyncJobsStore.created >= min_date

    if filter.max_date:
        max_date = date_filter(filter.max_date, 'max')
        filter_max_date = AsyncJobsStore.created <= max_date

    filter_service = text('1 = 1')  # (default) always true expression
    if filter.service:
        filter_service = AsyncJobsStore.service == filter.service
    filter_completed = text('1 = 1')  # (default) always true expression
    if filter.completed is not None:
        filter_completed = (
            AsyncJobsStore.completed.is_not(None)
            if filter.completed
            else AsyncJobsStore.completed.is_(None)
        )

    with Session(db_connection()) as session:
        query = get_jobs_query(
            session=session,
            filter_min_date=filter_min_date,
            filter_max_date=filter_max_date,
            filter_service=filter_service,
            filter_completed=filter_completed,
        )
        result = query_data_pagination(
            query=query,
            page=filter.page,
            page_size=filter.page_size
        )
        return result


def get_jobs_query(
    session: Session,
    filter_min_date: BinaryExpression,
    filter_max_date: BinaryExpression,
    filter_service: BinaryExpression,
    filter_completed: BinaryExpression,
) -> Query:
    """
    Constructs a SQLAlchemy query to retrieve async jobs based on specified filters.
    """

    query = (
        session.query(
            AsyncJobsStore.uuid,
            AsyncJobsStore.service,
            AsyncJobsStore.payload,
            AsyncJobsStore.expires,
            AsyncJobsStore.created,
            AsyncJobsStore.completed,
            AsyncJobsStore.locked_until,
            AsyncJobsStore.max_attempts,
            AsyncJobsStore.result,
        )
        .filter(filter_min_date, filter_max_date, filter_service, filter_completed)
        .order_by(desc(AsyncJobsStore.created))
    )
    return query


def create_job(
    service: str,
    payload: dict,
) -> AsyncJobsStore:
    """ Create async job """
    max_duration = payload.get('max_duration', MAX_DURATION)  # max duration in minutes
    max_attempts = payload.get('max_attempts', MAX_ATTEMPTS)
    tstamp = int(time.time()) + (max_duration * max_attempts * 2 * 60)
    expires = datetime.fromtimestamp(tstamp)

    with Session(db_connection()) as session:
        job_store = AsyncJobsStore(
            service=service,
            payload=payload,
            expires=expires,
            max_attempts=max_attempts,
        )
        session.expire_on_commit = False
        session.add(job_store)
        session.commit()

        return job_store


def complete_job(
    uuid: str,
    result: dict,
) -> None:
    """ Complete async job with result """
    # pylint: disable=broad-exception-caught
    log = logging.getLogger(__name__)
    job_store = single_job(uuid)
    if not job_store:
        log.error("Job %s not found", uuid)
        return
    now = datetime.now()
    if job_store.expires and job_store.expires < now:
        log.warning("Job %s is expired at %s", uuid, job_store.expires)
        return

    try:
        save_job_result(job_store=job_store, result=result)
    except Exception as exc:
        msg = exc.args[0]
        log.error("Error completing job - %s", msg)
        result = {'error': msg}
        try:
            save_job_result(job_store=job_store, result=result, error=True)
        except Exception as ex2:
            log.error("Error saving job error - %s", ex2.args[0])


def save_job_result(job_store: AsyncJobsStore, result: dict, error: bool = False):
    """Save Job result"""
    with Session(db_connection()) as session:
        job_store.completed = datetime.now()
        job_store.result = result
        if error:
            job_store.max_attempts = max(job_store.max_attempts - 1, 0)
        session.add(job_store)
        session.expire_on_commit = False
        session.commit()


def create_service(service: str) -> BaseService:
    """ Create job service from string """
    service_class = load_type(service, BaseService)

    return service_class()


def run_job_service(
    uuid: str,
) -> None:
    """ Create run job service """
    log = logging.getLogger(__name__)
    job_store = single_job(uuid)
    if not job_store:
        log.error("Job %s not found", uuid)
        return
    result = {}

    try:
        lock_job_service(job_store)
        service = create_service(job_store.service)
        job_store.payload['job_id'] = str(job_store.uuid)
        result = service.run(job_store.payload)

    except Exception as exc:  # pylint: disable=broad-exception-caught
        msg = f'{type(exc).__name__}: {exc}'
        log.error('Error in job service %s: %s', job_store.service, msg)
        result = {'error': msg}

    finally:
        complete_job(uuid=uuid, result=result)


def lock_job_service(
    job_store: AsyncJobsStore
) -> None:
    """Lock job service"""
    if not is_job_available(job_store):
        raise RuntimeError(f'Job {job_store.uuid} is not available')
    payload = job_store.payload if job_store.payload else {}
    tstamp = int(time.time()) + (int(payload.get('max_duration', MAX_DURATION)) * 60)
    locked_until = datetime.fromtimestamp(tstamp)

    with Session(db_connection()) as session:
        job_store.locked_until = locked_until
        session.add(job_store)
        session.expire_on_commit = False
        session.commit()


def is_job_available(
    job_store: AsyncJobsStore
) -> bool:
    """Check if job is available"""
    now = datetime.now()
    if job_store.completed or (job_store.expires and job_store.expires < now):
        return False
    if job_store.locked_until and (job_store.locked_until > now):
        return False
    if job_store.max_attempts <= 0:
        return False

    return True



"""Dates utilities"""
from datetime import datetime, time


def date_filter(date_str: str | None, type_str: str = 'min'):
    """
    Parses a date string into a datetime object with combined time information.

    Args:
        date_str (str): A string representing a date in the format 'YYYY-MM-DD'.
            None if no specific date is provided.

        type_str (str): Indicates whether to create a maximum or minimum date filter.
            Valid values are 'max' or 'min'. If None 'min' is i
    """
    max_date = datetime.now()
    min_date = datetime.fromtimestamp(0)

    if date_str is not None:
        parsed_date = datetime.strptime(date_str, '%Y-%m-%d')
        if type_str == 'max':
            max_date = parsed_date
            return datetime.combine(max_date, time.max)
        min_date = parsed_date
        return datetime.combine(min_date, time.min)

    return max_date if type_str == 'max' else min_date



"""Base class for text analysis services"""
from langchain.docstore.document import Document
from langchain.chains.summarize import load_summarize_chain
from langchain.chains.combine_documents.refine import RefineDocumentsChain
from langchain.chains.llm import LLMChain
from langchain_core.callbacks import Callbacks
from langchain_core.language_models import BaseLanguageModel
from langchain_core.prompts.loading import load_prompt_from_config
from brevia.callback import LoggingCallbackHandler
from brevia.load_file import read
from brevia.index import split_document
from brevia.models import load_chatmodel
from brevia.prompts import (
    load_prompt_from_yaml,
    load_stuff_prompts,
    load_map_prompts,
    load_refine_prompts
)
from brevia.settings import get_settings
from brevia.tasks.base import BaseAnalysisTask


class BaseTextAnalysisTask(BaseAnalysisTask):
    """Base class for text analysis tasks"""

    def text_documents(
            self,
            input_data: str,
            is_file_path: bool = True,
            options: dict | None = None
    ) -> list[Document]:
        """
        Load text from a file or directly from a string and split into documents.

        Args:
            input_data (str): The file path or the text content.
            is_file_path (bool): Flag to indicate if input_data is a file path
                or text content.
            options (dict | None): Additional options for splitting the document.

        Returns:
            list[Document]: A list of Document objects.
        """
        if is_file_path:
            text = read(file_path=input_data)
        else:
            text = input_data

        options = options or {}
        return split_document(Document(page_content=text), collection_meta=options)


class SummarizeTextAnalysisTask(BaseTextAnalysisTask):
    """Text summarization chain"""

    def __init__(
        self,
        text: str,
        chain_type: str | None = None,
        initial_prompt: dict | None = None,
        iteration_prompt: dict | None = None,
        llm_conf: dict | None = None,
        text_options: dict | None = None,
    ):
        self.text = text
        self.chain_type = chain_type
        self.initial_prompt = initial_prompt
        self.iteration_prompt = iteration_prompt
        self.llm_conf = llm_conf
        self.text_options = text_options
        self.load_analysis_prompts({
            'initial_prompt': self.initial_prompt,
            'iteration_prompt': self.iteration_prompt
        })

    def perform_task(self):
        """Service logic"""

        text_documents = self.text_documents(
            self.text, is_file_path=False, options=self.text_options)

        result = self.run_summarize_chain(text_documents)

        return {
            'input_documents': [{
                'page_content': doc.page_content,
                'metadata': doc.metadata
            } for doc in text_documents
            ],
            'output_text': result.get('output_text')
        }

    def load_analysis_prompts(self, prompts: dict | None = None):
        """Load analysis prompts"""
        if not prompts:
            raise ValueError('Prompts dictionary must be provided.')

        summarize_chain_map = {
            'stuff': load_stuff_prompts,
            'map_reduce': load_map_prompts,
            'refine': load_refine_prompts
        }

        chain_type = self.chain_type or get_settings().summ_default_chain
        if chain_type not in summarize_chain_map:
            raise ValueError(
                f"Got unsupported chain type: {chain_type}. "
                f"Should be one of {list(summarize_chain_map.keys())}"
            )

        missing = {'initial_prompt'} - prompts.keys()
        if missing:
            raise ValueError(f'Missing required prompts: {", ".join(missing)}')

        # load chain_type specific prompts:
        self.prompts = summarize_chain_map[chain_type](prompts)
        self.chain_type = chain_type

    def run_summarize_chain(
        self,
        pages: list[Document],
    ) -> dict[str, str]:
        """Run summarization chain"""

        logging_handler = LoggingCallbackHandler()
        llm_conf = self.llm_conf or get_settings().summarize_llm.copy() \
            # pylint: disable=no-member
        llm_conf['callbacks'] = [logging_handler]
        llm_text = load_chatmodel(llm_conf)

        kwargs = {
            'llm': llm_text,
            'chain_type': self.chain_type,
            'verbose': get_settings().verbose_mode,
            'callbacks': [logging_handler],
        }

        chain = load_summarize_chain(**kwargs, **self.prompts)
        return chain.invoke({'input_documents': pages}, return_only_outputs=True)


class RefineTextAnalysisTask(BaseTextAnalysisTask):
    """Text analysis using refine chain"""

    def __init__(
        self,
        file_path: str,
        prompts: dict | None = None,
        llm_conf: dict | None = None,
        text_options: dict | None = None,
    ):
        self.file_path = file_path
        self.llm_conf = llm_conf
        self.text_options = text_options
        self.load_analysis_prompts(prompts)

    def perform_task(self):
        """Service logic"""
        text_documents = self.text_documents(
            self.file_path, is_file_path=True, options=self.text_options)

        result = self.run_refine_chain(text_documents)

        return {
            'input_documents': [{
                'page_content': doc.page_content,
                'metadata': doc.metadata
            } for doc in text_documents
            ],
            'output_text': result.get('output_text')
        }

    def load_analysis_prompts(self, prompts: dict | None = None):
        """Load analysis prompts"""
        if not prompts:
            raise ValueError('Prompts dictionary must be provided.')
        missing = {'initial_prompt', 'refine_prompt'} - prompts.keys()
        if missing:
            raise ValueError(f'Missing required prompts: {", ".join(missing)}')

        for key, value in prompts.items():
            if isinstance(value, dict):
                prompts[key] = load_prompt_from_config(value)
            else:
                prompts[key] = load_prompt_from_yaml(value)

        self.prompts = prompts

    def run_refine_chain(
        self,
        pages: list[Document],
    ) -> dict[str, str]:
        """Run analysis with refine chain"""
        logging_handler = LoggingCallbackHandler()
        llm_conf = self.llm_conf or get_settings().summarize_llm.copy()  \
            # pylint: disable=no-member
        llm_conf['callbacks'] = [logging_handler]
        llm_text = load_chatmodel(llm_conf)

        # Using refine algorithm with modified prompts
        # for informations extraction on multiple pages
        chain = self.load_refine_chain(
            llm=llm_text,
            callbacks=[logging_handler],
        )

        return chain.invoke({'input_documents': pages}, return_only_outputs=True)

    def load_refine_chain(
        self,
        llm: BaseLanguageModel,
        callbacks: Callbacks,
        verbose: bool = True,
    ) -> RefineDocumentsChain:
        """Load refine chain"""
        initial_prompt = self.prompts.get('initial_prompt')
        refine_prompt = self.prompts.get('refine_prompt')

        return RefineDocumentsChain(
            initial_llm_chain=LLMChain(llm=llm, prompt=initial_prompt, verbose=verbose),
            refine_llm_chain=LLMChain(llm=llm, prompt=refine_prompt, verbose=verbose),
            document_variable_name='text',
            initial_response_name='existing_answer',
            verbose=verbose,
            callbacks=callbacks,
        )



"""Index document with embeddings in vector database."""
from functools import lru_cache
from os import path
from logging import getLogger
from warnings import warn
from langchain_community.vectorstores.pgembedding import CollectionStore
from langchain_community.vectorstores.pgembedding import EmbeddingStore
from langchain_community.vectorstores.pgvector import PGVector
from langchain_core.documents import Document
from langchain_text_splitters import NLTKTextSplitter
from langchain_text_splitters.base import TextSplitter
from requests import HTTPError
from sqlalchemy.orm import Session
from brevia import connection, load_file
from brevia.collections import single_collection_by_name
from brevia.models import load_embeddings
from brevia.settings import get_settings
from brevia.utilities.json_api import query_data_pagination
from brevia.utilities.types import load_type


def init_index():
    """Init index data"""
    warn("init_index deprecated, use init_splitting_data instead", DeprecationWarning)
    init_splitting_data()


@lru_cache
def init_splitting_data() -> bool:
    """Init splitting tools data (NLTK for now)"""
    try:
        import nltk  # pylint: disable=import-outside-toplevel
        return nltk.download('punkt_tab')

    except ImportError as exc:
        raise ImportError(
            "NLTK is not installed!"
        ) from exc


def load_pdf_file(
    # pylint: disable=too-many-arguments
    file_path: str,
    collection_name: str,
    document_id: str | None = None,
    metadata: dict | None = None,
    page_from: int = None,  # from page, used in PDF
    page_to: int = None,  # to page, used in PDF
) -> int:
    """
    Load document from PDF file, add document to collection index
    and return number of splitted text chunks
    """
    if not path.isfile(file_path):
        raise FileNotFoundError(file_path)

    text = load_file.read_pdf_file(
        file_path=file_path,
        page_from=page_from,
        page_to=page_to,
    )
    if metadata is None:
        metadata = {'source': path.basename(file_path)}

    return add_document(
        document=Document(
            page_content=text,
            metadata=metadata,
        ),
        collection_name=collection_name,
        document_id=document_id,
    )


def add_document(
    document: Document,
    collection_name: str,
    document_id: str = None,
) -> int:
    """ Add document to index and return number of splitted text chunks"""
    collection = single_collection_by_name(collection_name)
    coll_meta = collection.cmetadata if collection and collection.cmetadata else {}
    embed_conf = coll_meta.get('embeddings', None)
    texts = split_document(
        document=document,
        collection_meta=coll_meta,
    )
    PGVector.from_documents(
        embedding=load_embeddings(embed_conf),
        documents=texts,
        collection_name=collection_name,
        connection_string=connection.connection_string(),
        ids=[document_id] * len(texts),
        use_jsonb=True,
    )

    return len(texts)


def split_document(
    document: Document, collection_meta: dict = {}
) -> list[Document]:
    """ Split document into text chunks and return a list of documents"""
    init_splitting_data()
    text_splitter = create_splitter(collection_meta)
    texts = text_splitter.split_documents([document])
    counter = 1
    for text in texts:
        text.metadata['part'] = counter
        counter += 1
    return texts


def create_splitter(collection_meta: dict) -> TextSplitter:
    """ Create text splitter"""
    settings = get_settings()
    custom_splitter = collection_meta.get(
        'text_splitter',
        settings.text_splitter.copy()
    )
    chunk_size = int(collection_meta.get('chunk_size', settings.text_chunk_size))
    chunk_overlap = int(
        collection_meta.get('chunk_overlap', settings.text_chunk_overlap)
    )

    if not custom_splitter:
        return NLTKTextSplitter(
            separator="\n",
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap
        )

    chunk_conf = {'chunk_size': chunk_size, 'chunk_overlap': chunk_overlap}

    return create_custom_splitter({**chunk_conf, **custom_splitter})


def create_custom_splitter(split_conf: dict) -> TextSplitter:
    """ Create custom text splitter"""
    splitter_name = split_conf.pop('splitter', '')
    splitter_class = load_type(splitter_name, TextSplitter)

    return splitter_class(**split_conf)


def remove_document(
    collection_id: str,
    document_id: str,
):
    """ Remove document `document_id` from collection index"""
    filter_document = EmbeddingStore.custom_id == document_id
    filter_collection = EmbeddingStore.collection_id == collection_id
    with Session(connection.db_connection()) as session:
        query = session.query(EmbeddingStore).filter(filter_collection, filter_document)
        query.delete()
        session.commit()


def read_document(
    collection_id: str,
    document_id: str,
):
    """ Read document `document_id` from collection index"""
    filter_document = EmbeddingStore.custom_id == document_id
    filter_collection = EmbeddingStore.collection_id == collection_id
    with Session(connection.db_connection()) as session:
        query = session.query(EmbeddingStore.document, EmbeddingStore.cmetadata)
        query = query.filter(filter_collection, filter_document)
        return [row._asdict() for row in query.all()]


def metadata_filters(filter: dict[str, str] = {}):
    """
        Returns a list of metadata to be use as query filter in embeddings table
    """
    res = []
    for key in filter.keys():
        res.append(EmbeddingStore.cmetadata[key].astext == str(filter[key]))
    return res


def collection_documents(
    collection_id: str,
    filter: dict[str, str] = {},
    page: int = 1,
    page_size: int = 50,
):
    """ Read document `document_id` from collection index"""
    query_filters = [EmbeddingStore.collection_id == collection_id]
    query_filters.extend(metadata_filters(filter=filter))
    with Session(connection.db_connection()) as session:
        query = session.query(
            EmbeddingStore.document,
            EmbeddingStore.cmetadata,
            EmbeddingStore.custom_id
        )
        query = query.filter(*query_filters)
        return query_data_pagination(query=query, page=page, page_size=page_size)


def update_metadata(
    collection_id: str,
    document_id: str,
    metadata: dict | None = None,
):
    """ Update metadata of a document in a collection"""
    filter_document = EmbeddingStore.custom_id == document_id
    filter_collection = EmbeddingStore.collection_id == collection_id
    with Session(connection.db_connection()) as session:
        query = session.query(EmbeddingStore).filter(filter_collection, filter_document)
        query.update({EmbeddingStore.cmetadata: metadata})
        session.commit()


def documents_metadata(
    collection_id: str,
    filter: dict[str, str] = {},
    document_id: str = None,
):
    """ Read documents metadata of a collection"""
    query_filters = [EmbeddingStore.collection_id == collection_id]
    if document_id:
        query_filters.append(EmbeddingStore.custom_id == document_id)
    query_filters.extend(metadata_filters(filter=filter))
    with Session(connection.db_connection()) as session:
        query = session.query(EmbeddingStore.custom_id, EmbeddingStore.cmetadata)
        query = query.filter(*query_filters)
        result = []
        docs = []
        for row in query.all():
            item = row._asdict()
            if item['custom_id'] not in docs:
                docs.append(item['custom_id'])
                result.append(item)

        return result


def update_links_documents(collection_name: str) -> int:
    """ Update links document contents of a collection, if changed"""
    log = getLogger(__file__)
    log.info('Updating links contents in "%s"', collection_name)
    collection = single_collection_by_name(collection_name)
    if collection is None:
        log.error('Collection "%s" not found', collection_name)
        return 0
    docs_meta = documents_metadata(collection_id=collection.uuid,
                                   filter={'type': 'links'})
    count = 0
    for doc in docs_meta:
        try:
            res = update_collection_link(
                collection=collection,
                document_id=doc['custom_id'],
                document_medatata=doc['cmetadata']
            )
        except HTTPError as exc:
            log.error('HTTP Error updating document "%s" - %s', doc['custom_id'], exc)
            doc['cmetadata'] |= {'http_error': str(exc.response.status_code)}
            update_metadata(collection_id=collection.uuid,
                            document_id=doc['custom_id'],
                            metadata=doc['cmetadata'])
            res = False
        count += 1 if res else 0

    return count


def update_collection_link(collection: CollectionStore, document_id: str,
                           document_medatata: dict) -> bool:
    """ Update a document link in a collection"""
    log = getLogger(__file__)
    url = document_medatata.get('url')
    if not url:
        log.error('Document "%s" has no URL metadata', document_id)
        return False

    options = select_load_link_options(
        url=url,
        options=collection.cmetadata.get('link_load_options', [])
    )
    text = load_file.read_html_url(url=url, **options)
    document = Document(page_content=text, metadata=document_medatata)
    if document_has_changed(document=document, collection_id=collection.uuid,
                            document_id=document_id):
        remove_document(collection_id=collection.uuid, document_id=document_id)
        document.metadata.pop('http_error', None)
        add_document(document=document, collection_name=collection.name,
                     document_id=document_id)
        return True

    return False


def document_has_changed(document: Document, collection_id: str, document_id: str):
    """ Check if a document has changed """
    stored_docs = read_document(collection_id=collection_id, document_id=document_id)
    stored_docs.sort(key=lambda x: x['cmetadata'].get('part', 0))

    splitted = split_document(document)
    if len(stored_docs) != len(splitted):
        return True
    for i, stored_doc in enumerate(stored_docs):
        if stored_doc['document'] != splitted[i].page_content:
            return True
        if stored_doc['cmetadata'] != splitted[i].metadata:
            return True

    return False


def select_load_link_options(url: str, options: list) -> dict:
    """ Select load link options for a given URL"""
    unique_keys = set()
    for option in options:
        unique_keys.update(option.keys())
    unique_keys.discard('url')

    res = {}
    for key in unique_keys:
        res[key] = load_link_option(url=url, name=key, options=options)

    return res


def load_link_option(url: str, name: str, options: list) -> str | None:
    """Find load link option for a given URL"""
    item = next((x for x in options if url == x['url']), None)
    if item and item.get(name) is not None:
        return item.get(name)

    item = next((x for x in options if url.startswith(x['url'])), None)
    return item.get(name) if item else None



"""Locked until column

Revision ID: 1d927e0cbc61
Revises: be67c72a4e5a
Create Date: 2023-08-08 12:52:50.336008

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '1d927e0cbc61'
down_revision = 'be67c72a4e5a'
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.add_column(
        'async_jobs',
        sa.Column(
            'locked_until',
            sa.TIMESTAMP(timezone=False),
            comment='Timestamp at which the lock expires',
        ),
    )
    op.add_column(
        'async_jobs',
        sa.Column(
            'max_attempts',
            sa.SMALLINT(),
            nullable=False,
            server_default='1',
            comment='Maximum number of attempts left for this job'
        ),
    )


def downgrade() -> None:
    op.drop_column('async_jobs', 'locked_until')
    op.drop_column('async_jobs', 'max_attempts')
